import functools
import math
import operator
from math import ceil
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List

from pytezos.rpc.errors import RpcError


class OperationResult:
    """Operation result representation + useful parsing helpers for operation group"""

    def __init__(self, **props):
        self.props = props
        for key, value in props.items():
            setattr(self, key, value)

    def __repr__(self):
        res = [
            super(OperationResult, self).__repr__(),
            '\nProperties',
            *list(map(lambda x: f'.{x}', self.props)),
        ]
        return '\n'.join(res)

    @staticmethod
    def iter_contents(operation_group: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        """ Lazily iterate operation group contents including internal operations.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: generator
        """
        contents = operation_group.get('contents', [operation_group])
        for content in contents:
            yield {'internal': False, **content}
            internal_operation_results = content.get('metadata', {}).get('internal_operation_results', [])
            for result in internal_operation_results:
                yield {'internal': True, **result}

    @staticmethod
    def iter_results(operation_group: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        """ Lazily iterate operation results including internal operation results.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: generator
        """
        for content in OperationResult.iter_contents(operation_group):
            if content['internal'] and content.get('result'):
                yield content['result']
            elif not content['internal'] and content.get('metadata', {}).get('operation_result'):
                yield content['metadata']['operation_result']

    @staticmethod
    def consumed_gas(operation_group: Dict[str, Any]) -> int:
        """ Get total consumed gas for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(
            map(
                lambda x: math.ceil(int(x.get('consumed_milligas', '0')) / 1000),
                OperationResult.iter_results(operation_group),
            ),
        )

    @staticmethod
    def paid_storage_size_diff(operation_group: Dict[str, Any]) -> int:
        """ Get total paid storage size diff for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(
            map(
                lambda x: int(x.get('paid_storage_size_diff', '0')),
                OperationResult.iter_results(operation_group),
            ),
        )

    @staticmethod
    def burned(operation_group: Dict[str, Any]) -> int:
        """ Get total burned (due to account allocations) for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(
            map(
                lambda x: 257 if x.get('allocated_destination_contract') or x.get('originated_contracts') else 0,
                OperationResult.iter_results(operation_group),
            )
        )

    @staticmethod
    def is_applied(operation_group: Dict[str, Any]) -> bool:
        """ Check if ALL operations in a group are applied.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return all(map(lambda x: x['status'] == 'applied', OperationResult.iter_results(operation_group)))

    @staticmethod
    def errors(operation_group: Dict[str, Any]) -> List[Dict[str, Any]]:
        """ Collect errors from all operation results in a group.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: list of errors [{"id": "", ...}]
        """
        all_errors = (
            result.get("errors", []) if result["status"] != "applied" else []
            for result in OperationResult.iter_results(operation_group)
        )
        return functools.reduce(operator.iconcat, all_errors, [])

    @staticmethod
    def originated_contracts(operation_group: Dict[str, Any]) -> List[str]:
        """ Collect originated contract addresses from all operation results in a group.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: list of addresses ["tz12345...", ...]
        """
        originated_contracts = []
        for result in OperationResult.iter_results(operation_group):
            originated_contracts.extend(result.get('originated_contracts', []))
        return originated_contracts

    @staticmethod
    def get_contents(operation_group: Dict[str, Any], **predicates) -> List[Dict[str, Any]]:
        def match(x):
            return all(map(lambda pred: x.get(pred[0]) == pred[1], predicates.items()))

        if not predicates:
            return operation_group['contents']
        else:
            return list(filter(match, OperationResult.iter_contents(operation_group)))

    @staticmethod
    def get_result(content: Dict[str, Any]) -> Dict[str, Any]:
        if content.get('metadata'):
            return content['metadata']['operation_result']
        elif content.get('result'):
            return content['result']
        else:
            raise AssertionError(content)

    @classmethod
    def from_operation_group(cls, operation_group: Dict[str, Any], **predicates) -> List['OperationResult']:
        """ Initialize with operation group contents.

        :param operation_group: operation_group: {"branch": "B...", "contents": [...], ...} \
        :param predicates: filter contents using predicates `field=value`
        :rtype: List[OperationResult]
        """
        if not cls.is_applied(operation_group):
            raise RpcError.from_errors(cls.errors(operation_group))

        def dispatch(content):
            if content['kind'] == 'transaction':
                return cls.from_transaction(content)
            elif content['kind'] == 'origination':
                return cls.from_origination(content)
            else:
                return content

        contents = cls.get_contents(operation_group, **predicates)
        return list(map(dispatch, contents))

    @classmethod
    def from_origination(cls, content: Dict[str, Any]) -> 'OperationResult':
        """Initialize with origination content.

        :param content:
        :rtype: OperationResult
        """
        operation_result = cls.get_result(content)
        return cls(
            storage=content['script']['storage'],
            originated_contracts=operation_result['originated_contracts'],
        )

    @classmethod
    def from_transaction(cls, content: Dict[str, Any]) -> 'OperationResult':
        """Initialize with transaction content.

        :param content:
        :rtype: OperationResult
        """
        operation_result = cls.get_result(content)
        return cls(
            parameters=content.get('parameters'),
            storage=operation_result.get('storage'),
            lazy_diff=operation_result.get('lazy_storage_diff', []),
            # TODO: if it is already an internal operation, we should think... (build a tree?)
            operations=cls.get_contents(content, source=content['destination']),
        )



from typing import List
from typing import Type
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.logging import logger
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.big_map import BigMapType


class StackEltInstruction(MichelsonInstruction, prim='Stack_elt', args_len=2):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        raise RuntimeError('`Stack_elt` primitive is used only in TZT tests and cannot be executed directly')

    @classmethod
    def push(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res_type: MichelsonType
        literal: Type[MichelineLiteral]
        res_type, literal = cls.args  # type: ignore
        if res_type.prim == 'big_map':
            if isinstance(literal.literal, int):
                res = context.tzt_big_maps[literal.literal]  # type: ignore
            else:
                res = res_type.from_literal(literal)
                res.attach_context(context)
        elif res_type.is_pushable():
            res = res_type.from_literal(literal)
        else:
            raise Exception(f'`{res_type.prim}` is neither pushable nor big_map')

        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore

    @classmethod
    def pull(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res_type: MichelsonType
        literal: Type[MichelineLiteral]
        res_type, literal = cls.args  # type: ignore
        res = res_type.from_literal(literal)
        if res_type.is_pushable():
            expected_res = stack.pop1()
        elif res_type.prim == 'big_map':
            if issubclass(literal, MichelineSequence):
                expected_res = stack.pop1()
            else:
                expected_res = stack.pop1()
                # NOTE: We care about validity of the pointer, not it's contents
                res = expected_res
        else:
            raise Exception(f'`{res_type.prim}` is neither pushable nor big_map')

        if res != expected_res:
            logger.debug('expected: %s(%s)', expected_res.__class__.__name__, expected_res.__dict__)
            logger.debug('actual: %s(%s)', res.__class__.__name__, res.__dict__)
            raise Exception('Stack content is not equal to expected')

        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore


class BigMapInstruction(MichelsonInstruction, prim='Big_map', args_len=4):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        raise RuntimeError('`Big_map` primitive is used only in TZT tests and cannot be executed directly')

    @classmethod
    def add(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res_type: MichelsonType
        key_type: Type[MichelsonType]
        value_type: Type[MichelsonType]
        literal: Type[MichelineLiteral]
        res_type, key_type, value_type, literal = cls.args  # type: ignore

        big_map = BigMapType.empty(key_type, value_type)
        big_map.ptr = cast(int, res_type.literal)
        big_map.attach_context(context)
        for arg in literal.args:
            _, big_map = big_map.update(  # type: ignore
                key=key_type.from_literal(arg.args[0]),
                val=value_type.from_literal(arg.args[1]),
            )
        context.tzt_big_maps[big_map.ptr] = big_map  # type: ignore

        stdout.append(format_stdout(cls.prim, [], [literal]))  # type: ignore



import functools
import re
from collections import namedtuple
from typing import Tuple

from pytezos.michelson.tags import prim_tags

COMPARE = {'prim': 'COMPARE'}
UNIT = {'prim': 'UNIT'}
FAILWITH = {'prim': 'FAILWITH'}
DUP = {'prim': 'DUP'}
SWAP = {'prim': 'SWAP'}
CAR = {'prim': 'CAR'}
CDR = {'prim': 'CDR'}
CAR__ = {'prim': 'CAR', 'annots': ['@%%']}
CDR__ = {'prim': 'CDR', 'annots': ['@%%']}
DROP = {'prim': 'DROP'}
FAIL = [[UNIT, FAILWITH]]

macros = []

PxrNode = namedtuple('PxrNode', ['depth', 'annots', 'args', 'is_root'])


def macro(regexp):
    def register_macro(func):
        macros.append((re.compile(regexp), func))

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return register_macro


def seq(instr=None) -> list:
    if instr is None:
        return []
    elif isinstance(instr, list):
        return instr
    else:
        return [instr]


def expand_macro(prim, annots, args, internal=False):
    """Expands Michelson macro.

    :param prim: macro name
    :param annots: annotations (optional)
    :param args: arguments (optional)
    :param internal: this function is called during another macro expansion
    :returns: Code sequence (Micheline expression)
    """
    assert isinstance(annots, list)
    assert isinstance(args, list)
    if prim in prim_tags:
        return expr(prim=prim, annots=annots, args=args)

    for regexp, handler in macros:
        groups = regexp.findall(prim)
        if groups:
            assert len(groups) == 1
            res = handler(groups[0], annots, args)
            return res if internal else seq(res)

    raise AssertionError(f'unknown primitive `{prim}`')


def get_field_annots(annots):
    return list(filter(lambda x: isinstance(x, str) and x[0] == '%', annots))


def get_var_annots(annots):
    return list(filter(lambda x: isinstance(x, str) and x[0] == '@', annots))


def skip_nones(array):
    return list(filter(lambda x: x is not None, array))


def expr(**kwargs) -> dict:
    return {k: v for k, v in kwargs.items() if v}


def dip_n(instr, depth=1):
    if depth <= 0:
        return instr
    elif depth == 1:
        return expr(
            prim='DIP',
            args=[seq(instr)],
        )
    else:
        return expr(
            prim='DIP',
            args=[{'int': str(depth)}, seq(instr)],
        )


@macro(r'^CMP(EQ|NEQ|LT|GT|LE|GE)$')
def expand_cmpx(prim, annots, args) -> list:
    assert not args
    return [
        COMPARE,
        expr(prim=prim, annots=annots),
    ]


@macro(r'^IF(EQ|NEQ|LT|GT|LE|GE)$')
def expand_ifx(prim, annots, args) -> list:
    assert len(args) == 2
    return [
        expr(prim=prim, annots=annots),
        expr(prim='IF', args=args),
    ]


@macro(r'^IFCMP(EQ|NEQ|LT|GT|LE|GE)$')
def expand_ifcmpx(prim, annots, args) -> list:
    assert len(args) == 2
    return [
        [COMPARE, expr(prim=prim, annots=annots)],
        expr(prim='IF', args=args),
    ]


@macro(r'^FAIL$')
def expand_fail(prim, annots, args) -> list:
    assert not annots
    assert not args
    return [
        UNIT,
        FAILWITH,
    ]


@macro(r'^ASSERT$')
def expand_assert(prim, annots, args) -> dict:
    assert not annots
    assert not args
    return expr(
        prim='IF',
        args=[[], FAIL],
    )


@macro(r'^ASSERT_(EQ|NEQ|LT|LE|GT|GE)$')
def expand_assert_x(prim, annots, args) -> list:
    assert not args
    assert not annots  # TODO: ask why
    return expand_ifx(
        prim,
        annots=[],
        args=[[], FAIL],
    )


@macro(r'^ASSERT_CMP(EQ|NEQ|LT|LE|GT|GE)$')
def expand_assert_cmpx(prim, annots, args) -> list:
    assert not args
    assert not annots  # TODO: ask why
    return expand_ifcmpx(
        prim,
        annots=[],
        args=[[], FAIL],
    )


@macro(r'^ASSERT_NONE$')
def expand_assert_none(prim, annots, args) -> dict:
    assert not annots
    assert not args
    return expr(
        prim='IF_NONE',
        args=[[], FAIL],
    )


@macro('^ASSERT_SOME$')
def expand_assert_some(prim, annots, args) -> dict:
    assert not args
    return expr(
        prim='IF_NONE',
        args=[FAIL, [expr(prim='RENAME', annots=annots)]],
    )


@macro('^ASSERT_LEFT$')
def expand_assert_left(prim, annots, args) -> dict:
    assert not args
    return expr(
        prim='IF_LEFT',
        args=[[expr(prim='RENAME', annots=annots)], FAIL],
    )


@macro('^ASSERT_RIGHT$')
def expand_assert_right(prim, annots, args) -> dict:
    assert not args
    return expr(
        prim='IF_LEFT',
        args=[FAIL, [expr(prim='RENAME', annots=annots)]],
    )


@macro(r'^D(II+)P$')
def expand_dixp(prim, annots, args) -> dict:
    assert not annots
    assert len(args) == 1
    return dip_n(args, depth=len(prim))


@macro(r'^D(UU+)P$')
def expand_duxp(prim, annots, args) -> dict:
    assert not args
    depth = len(prim)
    return expr(
        prim='DUP',
        annots=annots,
        args=[{'int': str(depth)}],
    )


def build_pxr_tree(pxr_macro, pxr_annots) -> PxrNode:
    def parse(prim, annots, depth=0, is_root=False):
        letter, prim = prim[0], prim[1:]
        if letter == 'P':
            dip_depth = depth
            left, l_annot, prim, annots, depth = parse(prim, annots, depth)
            right, r_annot, prim, annots, depth = parse(prim, annots, depth)
            return (
                PxrNode(
                    dip_depth,
                    [l_annot, r_annot],
                    [left, right],
                    is_root,
                ),
                None,
                prim,
                annots,
                depth,
            )
        else:
            annot, annots = (annots[0], annots[1:]) if annots else (None, [])
            return letter, annot, prim, annots, depth + 1

    root, _, _, _, _ = parse(pxr_macro, pxr_annots, is_root=True)
    return root


def traverse_pxr_tree(prim, annots, produce):
    res = []

    def walk(node):
        if isinstance(node, PxrNode):
            res.insert(0, dip_n(produce(node), depth=node.depth))
            _ = list(map(walk, node.args))

    walk(build_pxr_tree(prim, annots))
    return res


@macro(r'^P[PAI]{3,}R$')
def expand_pxr(prim, annots, args) -> list:
    def produce(node: PxrNode):
        pair_annots = [node.annots[0] or '%', node.annots[1]] if any(node.annots) else []
        if node.is_root:
            pair_annots.extend(get_var_annots(annots))
        return expr(prim='PAIR', annots=skip_nones(pair_annots))

    assert not args
    return traverse_pxr_tree(prim, get_field_annots(annots), produce)


@macro(r'^UN(P[PAI]{3,}R)$')
def expand_unpxr(prim, annots, args) -> list:
    def produce(node: PxrNode):
        return [expr(prim='UNPAIR', annots=skip_nones(node.annots))]

    assert not args
    return list(reversed(traverse_pxr_tree(prim, annots, produce)))


def expand_cxr(prim, annots) -> list:
    return seq(expand_macro(prim=f'C{prim}R', annots=annots, args=[], internal=True))


@macro(r'^CA([AD]+)R$')
def expand_caxr(prim, annots, args) -> list:
    assert not args
    return [
        CAR,
        *expand_cxr(prim, annots),
    ]


@macro(r'^CD([AD]+)R$')
def expand_cdxr(prim, annots, args) -> list:
    assert not args
    return [
        CDR,
        *expand_cxr(prim, annots),
    ]


@macro(r'^IF_SOME$')
def expand_if_some(prim, annots, args) -> dict:
    assert not annots
    assert len(args) == 2
    return expr(
        prim='IF_NONE',
        args=list(reversed(args)),
    )


@macro(r'^IF_RIGHT$')
def expand_if_right(prim, annots, args) -> dict:
    assert not annots
    assert len(args) == 2
    return expr(
        prim='IF_LEFT',
        args=list(reversed(args)),
    )


@macro(r'^SET_CAR$')
def expand_set_car(prim, annots, args) -> list:
    assert not args
    return [
        SWAP,
        expr(
            prim='UPDATE',
            args=[{'int': '1'}],
            annots=annots,
        ),
    ]


@macro(r'^SET_CDR$')
def expand_set_cdr(prim, annots, args) -> list:
    assert not args
    return [
        SWAP,
        expr(
            prim='UPDATE',
            args=[{'int': '2'}],
            annots=annots,
        ),
    ]


def expand_set_cxr(prim, annots):
    set_cxr = expand_macro(
        prim=f'SET_C{prim}R',
        annots=get_field_annots(annots),
        args=[],
        internal=True,
    )
    pair = expr(
        prim='PAIR',
        annots=['%@', '%@'] + get_var_annots(annots),
    )
    return set_cxr, pair


@macro(r'^SET_CA([AD]+)R$')
def expand_set_caxr(prim, annots, args) -> list:
    assert not args
    set_cxr, pair = expand_set_cxr(prim, annots)
    return [
        DUP,
        dip_n([CAR__, set_cxr]),
        CDR__,
        SWAP,
        pair,
    ]


@macro(r'^SET_CD([AD]+)R$')
def expand_set_cdxr(prim, annots, args) -> list:
    assert not args
    set_cxr, pair = expand_set_cxr(prim, annots)
    return [
        DUP,
        dip_n([CDR__, set_cxr]),
        CAR__,
        pair,
    ]


def get_map_cxr_annots(annots) -> Tuple[str, list]:
    field_annots = get_field_annots(annots)
    if field_annots:
        assert len(field_annots) == 1
        return field_annots[0], [f'@{field_annots[0][1:]}']
    else:
        return '%', []


@macro(r'^MAP_CAR$')
def expand_map_car(prim, annots, args) -> list:
    car_annot, var_annots = get_map_cxr_annots(annots)
    return [
        DUP,
        CDR__,
        dip_n(
            [
                expr(
                    prim='CAR',
                    annots=var_annots,
                ),
                *args,
            ]
        ),
        SWAP,
        expr(
            prim='PAIR',
            annots=[car_annot, '%@'],
        ),
    ]


@macro(r'^MAP_CDR$')
def expand_map_cdr(prim, annots, args) -> list:
    cdr_annot, var_annots = get_map_cxr_annots(annots)
    return [
        DUP,
        expr(prim='CDR', annots=var_annots),
        *args,
        SWAP,
        CAR__,
        expr(
            prim='PAIR',
            annots=['%@', cdr_annot],
        ),
    ]


def expand_map_cxr(prim, annots, args):
    set_cxr = expand_macro(
        prim=f'MAP_C{prim}R',
        annots=get_field_annots(annots),
        args=args,
        internal=True,
    )
    pair = expr(
        prim='PAIR',
        annots=['%@', '%@'] + get_var_annots(annots),
    )
    return set_cxr, pair


@macro(r'^MAP_CA([AD]+)R$')
def expand_map_caxr(prim, annots, args) -> list:
    map_cxr, pair = expand_map_cxr(prim, annots, args)
    return [
        DUP,
        dip_n([CAR__, map_cxr]),
        CDR__,
        SWAP,
        pair,
    ]


@macro(r'^MAP_CD([AD]+)R$')
def expand_map_cdxr(prim, annots, args) -> list:
    map_cxr, pair = expand_map_cxr(prim, annots, args)
    return [
        DUP,
        dip_n([CDR__, map_cxr]),
        CAR__,
        pair,
    ]



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

from pytezos.context.impl import ExecutionContext
from pytezos.crypto.encoding import base58_encode
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.instructions.tzt import BigMapInstruction
from pytezos.michelson.instructions.tzt import StackEltInstruction
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import get_script_section
from pytezos.michelson.micheline import get_script_sections
from pytezos.michelson.micheline import try_catch
from pytezos.michelson.micheline import validate_sections
from pytezos.michelson.sections.code import CodeSection
from pytezos.michelson.sections.parameter import ParameterSection
from pytezos.michelson.sections.storage import StorageSection
from pytezos.michelson.sections.tzt import AmountSection
from pytezos.michelson.sections.tzt import BalanceSection
from pytezos.michelson.sections.tzt import BigMapsSection
from pytezos.michelson.sections.tzt import ChainIdSection
from pytezos.michelson.sections.tzt import InputSection
from pytezos.michelson.sections.tzt import NowSection
from pytezos.michelson.sections.tzt import OutputSection
from pytezos.michelson.sections.tzt import SelfSection
from pytezos.michelson.sections.tzt import SenderSection
from pytezos.michelson.sections.tzt import SourceSection
from pytezos.michelson.sections.view import ViewSection
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import ListType
from pytezos.michelson.types import MichelsonType
from pytezos.michelson.types import OperationType
from pytezos.michelson.types import PairType


class MichelsonProgram:
    """Michelson .tz contract interpreter interface"""

    parameter: Type[ParameterSection]
    storage: Type[StorageSection]
    code: Type[CodeSection]
    views: List[Type[ViewSection]]

    def __init__(self, name: str, parameter: ParameterSection, storage: StorageSection) -> None:
        self.name = name
        self.parameter_value = parameter
        self.storage_value = storage

    @staticmethod
    def load(context: ExecutionContext, with_code=False) -> Type['MichelsonProgram']:
        """Create MichelsonProgram type from filled context"""
        cls = type(
            MichelsonProgram.__name__,
            (MichelsonProgram,),
            {
                'parameter': ParameterSection.match(context.get_parameter_expr()),
                'storage': StorageSection.match(context.get_storage_expr()),
                'code': CodeSection.match(context.get_code_expr() if with_code else []),
                'views': [ViewSection.match(expr) for expr in context.get_views_expr()] if with_code else [],
            },
        )
        return cast(Type['MichelsonProgram'], cls)

    @staticmethod
    def create(sequence: Type[MichelineSequence]) -> Type['MichelsonProgram']:
        """Create MichelsonProgram type from micheline"""
        validate_sections(
            sequence,
            (
                'parameter',
                'storage',
                'code',
            ),
        )
        cls = type(
            MichelsonProgram.__name__,
            (MichelsonProgram,),
            {
                'parameter': get_script_section(sequence, cls=ParameterSection, required=True),  # type: ignore
                'storage': get_script_section(sequence, cls=StorageSection, required=True),  # type: ignore
                'code': get_script_section(sequence, cls=CodeSection, required=True),  # type: ignore
                'views': get_script_sections(sequence, cls=ViewSection),  # type: ignore
            },
        )
        return cast(Type['MichelsonProgram'], cls)

    @staticmethod
    def match(expr) -> Type['MichelsonProgram']:
        seq = cast(Type[MichelineSequence], MichelineSequence.match(expr))
        if not issubclass(seq, MichelineSequence):
            raise Exception(f'Expected sequence, got {seq.prim}')
        return MichelsonProgram.create(seq)

    @classmethod
    def as_micheline_expr(cls) -> List[Dict[str, Any]]:
        return [
            cls.parameter.as_micheline_expr(),
            cls.storage.as_micheline_expr(),
            cls.code.as_micheline_expr(),
            *[view.as_micheline_expr() for view in cls.views],
        ]

    @classmethod
    def get_view(cls, name: str) -> Type[ViewSection]:
        return next(view for view in cls.views if view.name == name)

    @classmethod
    def instantiate(cls, entrypoint: str, parameter, storage) -> 'MichelsonProgram':
        parameter_value = cls.parameter.from_parameters({'entrypoint': entrypoint, 'value': parameter})
        storage_value = cls.storage.from_micheline_value(storage)
        return cls(entrypoint, parameter_value, storage_value)

    @classmethod
    def instantiate_view(cls, name: str, parameter, storage) -> 'MichelsonProgram':
        view = cls.get_view(name)
        parameter_ty = ParameterSection.create_type(args=[view.args[1]])
        parameter_value = parameter_ty.from_micheline_value(parameter)
        storage_value = cls.storage.from_micheline_value(storage)
        return cls(name, parameter_value, storage_value)

    @try_catch('BEGIN')
    def begin(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> None:
        """Prepare stack for contract execution"""
        self.parameter_value.attach_context(context)
        self.storage_value.attach_context(context)
        res = PairType.from_comb([self.parameter_value.item, self.storage_value.item])
        stack.push(res)
        stdout.append(format_stdout(f'BEGIN %{self.name}', [], [res]))

    def execute(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> MichelsonInstruction:
        """Execute contract in interpreter"""
        return cast(MichelsonInstruction, self.code.args[0].execute(stack, stdout, context))

    def execute_view(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext):
        """Execute view in interpreter"""
        view = self.get_view(self.name)
        return cast(MichelsonInstruction, view.args[3].execute(stack, stdout, context))

    @try_catch('END')
    def end(
        self,
        stack: MichelsonStack,
        stdout: List[str],
        output_mode='readable',
    ) -> Tuple[List[dict], Any, List[dict], PairType]:
        """Finish contract execution"""
        res = cast(PairType, stack.pop1())
        if len(stack):
            raise Exception(f'Stack is not empty: {repr(stack)}')
        res.assert_type_equal(
            PairType.create_type(
                args=[ListType.create_type(args=[OperationType]), self.storage.args[0]],
            ),
            message='list of operations + resulting storage',
        )
        operations = [op.content for op in res.items[0]]  # type: ignore
        lazy_diff = []  # type: ignore
        storage = res.items[1].aggregate_lazy_diff(lazy_diff).to_micheline_value(mode=output_mode)
        stdout.append(format_stdout(f'END %{self.name}', [res], []))
        return operations, storage, lazy_diff, res

    @try_catch('RET')
    def ret(
        self,
        stack: MichelsonStack,
        stdout: List[str],
        output_mode='readable',
    ) -> MichelsonType:
        view = self.get_view(self.name)
        res = stack.pop1()
        if len(stack):
            raise Exception(f'Stack is not empty: {repr(stack)}')
        res.assert_type_equal(view.args[2], message='view return type')
        stdout.append(format_stdout(f'RET %{self.name}', [res], []))
        return view.args[2].from_micheline_value(res.to_micheline_value(mode=output_mode))


class TztMichelsonProgram:
    """Michelson .tzt contract interpreter interface"""

    code: Type[CodeSection]
    input: Type[InputSection]
    output: Type[OutputSection]
    big_maps: Optional[Type[BigMapsSection]]

    @staticmethod
    def load(context: ExecutionContext, with_code=False):
        """Create TztMichelsonProgram type from filled context"""
        cls = type(
            TztMichelsonProgram.__name__,
            (TztMichelsonProgram,),
            {
                'input': InputSection.match(context.get_input_expr()),
                'output': OutputSection.match(context.get_output_expr()),
                'code': CodeSection.match(context.get_code_expr() if with_code else []),
                'big_maps': BigMapsSection.match(context.get_big_maps_expr()) if context.get_big_maps_expr() else None,
            },
        )
        return cast(Type['TztMichelsonProgram'], cls)

    @staticmethod
    def create(sequence: Type[MichelineSequence]) -> Type['TztMichelsonProgram']:
        """Create TztMichelsonProgram type from micheline"""
        validate_sections(sequence, ('input', 'output', 'code'))
        cls = type(
            TztMichelsonProgram.__name__,
            (TztMichelsonProgram,),
            {
                'input': get_script_section(sequence, cls=InputSection, required=True),  # type: ignore
                'output': get_script_section(sequence, cls=OutputSection, required=True),  # type: ignore
                'code': get_script_section(sequence, cls=CodeSection, required=True),  # type: ignore
                'big_maps': get_script_section(sequence, cls=BigMapsSection, required=False),  # type: ignore
            },
        )
        return cast(Type['TztMichelsonProgram'], cls)

    @staticmethod
    def match(expr) -> Type['TztMichelsonProgram']:
        seq = cast(Type[MichelineSequence], MichelineSequence.match(expr))
        if not issubclass(seq, MichelineSequence):
            raise Exception(f'expected sequence, got {seq.prim}')
        return TztMichelsonProgram.create(seq)

    @classmethod
    def as_micheline_expr(cls) -> List[Dict[str, Any]]:
        # TODO: Serialize all sections
        return [
            cls.code.as_micheline_expr(),
            cls.input.as_micheline_expr(),
            cls.output.as_micheline_expr(),
        ]

    @classmethod
    def instantiate(cls) -> 'TztMichelsonProgram':
        return cls()

    def fill_context(self, script, context: ExecutionContext) -> None:
        sender = context.get_sender_expr()
        if sender:
            context.sender = SenderSection.match(sender).args[0].get_string()  # type: ignore
        amount = context.get_amount_expr()
        if amount:
            context.amount = AmountSection.match(amount).args[0].get_int()  # type: ignore
        balance = context.get_balance_expr()
        if balance:
            context.balance = BalanceSection.match(balance).args[0].get_int()  # type: ignore
        _self = context.get_self_expr()
        if _self:
            context.address = SelfSection.match(_self).args[0].get_string()  # type: ignore
        now = context.get_now_expr()
        if now:
            context.now = NowSection.match(now).args[0].get_int()  # type: ignore
        source = context.get_source_expr()
        if source:
            context.source = SourceSection.match(source).args[0].get_string()  # type: ignore
        chain_id = context.get_chain_id_expr()
        if chain_id:
            # FIXME: Move to some common place
            context.chain_id = base58_encode(
                cast(bytes, ChainIdSection.match(chain_id).args[0].literal),
                prefix=b'Net',
            ).decode()

    def register_bigmaps(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> None:
        if self.big_maps:
            for item in self.big_maps.args[0].args[::-1]:
                if not issubclass(item, BigMapInstruction):
                    raise Exception('Only `Big_map` instructions can be used in `big_maps` section')
                item.add(stack, stdout, context)

    def begin(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> None:
        """Prepare stack for contract execution"""

        for item in self.input.args[0].args[::-1]:
            if issubclass(item, StackEltInstruction):
                item.push(stack, stdout, context)
            else:
                raise Exception('Only `Stack_elt` instructions can be used in `input` section', item)

    def execute(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> MichelsonInstruction:
        """Execute contract in interpreter"""
        return cast(MichelsonInstruction, self.code.args[0].execute(stack, stdout, context))

    def end(self, stack: MichelsonStack, stdout: List[str], context: ExecutionContext) -> None:
        """Finish contract execution"""
        for item in self.output.args[0].args:
            if not issubclass(item, StackEltInstruction):
                raise Exception('Only `Stack_elt` instructions can be used in `output` section')
            item.pull(stack, stdout, context)

        if len(stack):
            raise Exception('Stack is not empty after processing `output` section')



from pytezos.rpc.node import RpcError


class ReorgError(Exception):
    def __init__(self, level, old_hash, new_hash):
        super().__init__(level, old_hash, new_hash)
        self.level = level
        self.old_hash = old_hash
        self.new_hash = new_hash

    def __repr__(self):
        return f'Reorg detected at level {self.level}: {self.old_hash} -> {self.new_hash}'


class MichelsonBadContractParameter(RpcError, error_id='michelson_v1.bad_contract_parameter'):
    """Either no parameter was supplied to a contract with a non-unit parameter type, a non-unit parameter was passed
    to an account, or a parameter was supplied of the wrong type
    """


class MichelsonBadReturn(RpcError, error_id='michelson_v1.bad_return'):
    """Unexpected stack at the end of a lambda or script"""


class MichelsonError(RpcError, error_id='michelson_v1'):
    """Catch all michelson_v1 errors"""


class TezArithmeticError(RpcError, error_id='tez'):
    """Catch all tez errors"""


class MichelsonScriptRejected(RpcError, error_id='script_rejected'):
    """A FAILWITH instruction was reached"""



from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import Undefined


def get_type_layout(
    flat_args: List[Tuple[str, Type[MichelsonType]]],
    infer_names: bool = False,
    entrypoints: bool = False,
) -> Tuple[Optional[Dict[str, str]], Optional[Dict[str, str]], Dict[int, str]]:
    reserved = set()
    path_to_key = {}
    for i, (bin_path, arg) in enumerate(flat_args):
        key = arg.field_name
        if key is None and not entrypoints:
            key = arg.type_name
        if key is not None and key not in reserved:
            reserved.add(key)
            path_to_key[bin_path] = key
        else:
            assert entrypoints is False, f'duplicate key {key}'
            path_to_key[bin_path] = f'{arg.prim}_{i}'

    idx_to_path = {i: path for i, path in enumerate(path_to_key)}
    if len(reserved) == 0 and infer_names is False and entrypoints is False:
        path_to_key = None  # type: ignore
        key_to_path = None
    else:
        key_to_path = {name: path for path, name in path_to_key.items()}
    return path_to_key, key_to_path, idx_to_path


class Nested:
    def __init__(self, *args):
        self.args = args

    def __getitem__(self, item: int):
        return self.args[item]


def wrap_or(obj, path) -> Nested:
    if len(path) == 0:
        return obj
    elif path[0] == '0':
        return Nested(wrap_or(obj, path[1:]), Undefined)
    elif path[0] == '1':
        return Nested(Undefined, wrap_or(obj, path[1:]))
    else:
        raise AssertionError(path)


def wrap_pair(obj: dict, path='') -> Nested:
    if all(not x.startswith(path) for x in obj):
        raise KeyError(path)

    items = [obj[subpath] if subpath in obj else wrap_pair(obj, subpath) for subpath in [path + '0', path + '1']]
    return Nested(*items)


def wrap_parameters(expr, path):
    if len(path) == 0:
        return expr
    elif path[0] == '0':
        return {'prim': 'Left', 'args': [wrap_parameters(expr, path[1:])]}
    elif path[0] == '1':
        return {'prim': 'Right', 'args': [wrap_parameters(expr, path[1:])]}
    else:
        raise AssertionError(path)


class ADTMixin:
    @classmethod
    def iter_type_args(
        cls,
        entrypoints: bool = False,
        path='',
    ) -> Generator[Tuple[str, Type[MichelsonType]], None, None]:
        raise NotImplementedError

    @classmethod
    def get_flat_args(
        cls,
        infer_names: bool = False,
        force_tuple: bool = False,
        entrypoints: bool = False,
    ) -> Union[Dict[str, Type[MichelsonType]], List[Type[MichelsonType]]]:
        flat_args = list(cls.iter_type_args(entrypoints=entrypoints))
        if force_tuple is False:
            path_to_key, _, _ = get_type_layout(flat_args, infer_names=infer_names, entrypoints=entrypoints)
            if isinstance(path_to_key, dict):
                return {path_to_key[path]: arg for path, arg in flat_args}
        return [arg for _, arg in flat_args]

    @classmethod
    def get_type_layout(
        cls,
        infer_names: bool = False,
        entrypoints: bool = False,
    ) -> Tuple[Optional[Dict[str, str]], Optional[Dict[str, str]], Dict[int, str]]:
        flat_args = list(cls.iter_type_args(entrypoints=entrypoints))
        return get_type_layout(flat_args, infer_names=infer_names, entrypoints=entrypoints)

    def iter_values(self, path='') -> Generator[Tuple[str, MichelsonType], None, None]:
        raise NotImplementedError

    def get_flat_values(
        self,
        infer_names: bool = False,
        force_tuple: bool = False,
        entrypoints: bool = False,
    ) -> Union[Dict[str, MichelsonType], List[MichelsonType]]:
        path_to_key, _, _ = self.get_type_layout(infer_names=infer_names, entrypoints=entrypoints)
        flat_values = list(self.iter_values())
        if force_tuple is False and isinstance(path_to_key, dict):
            return {path_to_key[path]: arg for path, arg in flat_values}
        else:
            return [arg for _, arg in flat_values]

    def get_value(self, key: Union[str, int], infer_names: bool = False):
        _, key_to_path, idx_to_path = self.get_type_layout(infer_names=infer_names)
        if isinstance(key, str):
            assert key_to_path, f'type is not named'
            path = key_to_path[key]
        elif isinstance(key, int):
            path = idx_to_path[key]
        else:
            raise AssertionError(f'expected string or int, got {key}')
        return next(v for p, v in self.iter_values() if p == path)



from os.path import join

import requests


def create_deployment(repo_slug, oauth_token, environment, ref='master'):
    return requests.post(
        url=join('https://api.github.com/repos', repo_slug, 'deployments'),
        headers={
            'Accept': 'application/vnd.github.ant-man-preview+json',
            'Authorization': f'token {oauth_token}',
        },
        json={
            'ref': ref,
            'environment': environment,
            'required_contexts': [],
        },  # bypass checking
    ).json()


def create_deployment_status(repo_slug, oauth_token, deployment_id, state, environment, environment_url=None):
    assert state in {'success', 'failure'}
    return requests.post(
        url=join('https://api.github.com/repos', repo_slug, 'deployments', str(deployment_id), 'statuses'),
        headers={
            'Accept': 'application/vnd.github.flash-preview+json',
            'Authorization': f'token {oauth_token}',
        },
        json={
            'state': state,
            'environment': environment,
            'environment_url': environment_url,
        },
    ).json()



import re
from contextlib import suppress
from typing import Dict
from typing import List
from typing import Type
from typing import cast

import strict_rfc3339  # type: ignore

from pytezos.context.abstract import AbstractContext
from pytezos.context.mixin import nodes
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.sections import ParameterSection
from pytezos.michelson.sections import StorageSection
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import ListType
from pytezos.michelson.types import OperationType
from pytezos.michelson.types import PairType
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.core import FalseLiteral
from pytezos.michelson.types.core import TrueLiteral
from pytezos.rpc.node import RpcMultiNode
from pytezos.rpc.node import RpcNode
from pytezos.rpc.shell import ShellQuery


class DumpAllInstruction(MichelsonInstruction, prim='DUMP'):
    def __init__(self, items: List[MichelsonType]):
        super().__init__(stack_items_added=len(items))
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        items = stack.items[:]
        stdout.append(f'DUMP => {items}')
        return cls(items)


class DumpInstruction(MichelsonInstruction, prim='DUMP', args_len=1):
    def __init__(self, items: List[MichelsonType]):
        super().__init__(stack_items_added=len(items))
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        literal: Type[MichelineLiteral] = cls.args[0]  # type: ignore
        count = cast(int, literal.literal)
        count = min(count, len(stack))
        items = stack.items[:count]
        stdout.append(f'DUMP => {items}')
        return cls(items)


class PrintInstruction(MichelsonInstruction, prim='PRINT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        literal: Type[MichelineLiteral] = cls.args[0]  # type: ignore

        template = literal.get_string()

        def format_stack_item(match):
            i = int(match.groups()[0])
            assert i < len(stack), f'requested {i}th element, got only {len(stack)} items'
            return repr(stack.items[i])

        message = re.sub(r'{(\d+)}', format_stack_item, template)
        stdout.append(message)
        return cls()


class DebugInstruction(MichelsonInstruction, prim='DEBUG', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        literal = cls.args[0]
        if issubclass(literal, (TrueLiteral, FalseLiteral)):
            debug = literal.literal
        else:
            debug = bool(literal.get_int())  # type: ignore

        context.debug = debug  # type: ignore
        return cls()


class DropAllInstruction(MichelsonInstruction, prim='DROP_ALL'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        stack.items = []
        return cls()


class BeginInstruction(MichelsonInstruction, prim='BEGIN', args_len=2):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        # FIXME: MichelsonProgram copypaste
        parameter_literal, storage_literal = cls.args  # type: ignore

        parameter_type_expr = context.get_parameter_expr()
        storage_type_expr = context.get_storage_expr()
        if parameter_type_expr is None:
            raise Exception('parameter type is not initialized')
        if storage_type_expr is None:
            raise Exception('storage type is not initialized')

        parameter_type = ParameterSection.match(parameter_type_expr)
        storage_type = StorageSection.match(storage_type_expr)
        parameter = parameter_type.from_micheline_value(parameter_literal.as_micheline_expr())
        storage = storage_type.from_micheline_value(storage_literal.as_micheline_expr())

        parameter.attach_context(context)
        storage.attach_context(context)
        res = PairType.from_comb([parameter.item, storage.item])
        stack.items = []
        stack.push(res)
        stdout.append(format_stdout(f'BEGIN %default', [], [res]))
        return cls(stack_items_added=1)


class CommitInstruction(MichelsonInstruction, prim='COMMIT'):
    def __init__(self, lazy_diff: List[Dict[str, str]], result, stack_items_added: int = 0) -> None:
        super().__init__(stack_items_added)
        self.lazy_diff = lazy_diff
        self.result = result

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        # FIXME: MichelsonProgram copypaste
        debug, context.debug = context.debug, False  # type: ignore

        res = cast(PairType, stack.pop1())
        if len(stack):
            raise Exception(f'Stack is not empty: {stack}')
        res.assert_type_equal(
            PairType.create_type(
                args=[
                    ListType.create_type(args=[OperationType]),
                    StorageSection.match(context.get_storage_expr()).args[0],
                ],
            ),
            message='list of operations + resulting storage',
        )
        operations = ListType(items=list(res.items[0]))  # type: ignore
        lazy_diff = []  # type: ignore
        storage = res.items[1].aggregate_lazy_diff(lazy_diff)
        stdout.append(format_stdout(f'END %default', [res], []))

        result = PairType.from_comb([operations, storage])
        context.debug = debug  # type: ignore
        return cls(lazy_diff=lazy_diff, result=result)


class RunInstruction(MichelsonInstruction, prim='RUN', args_len=3):
    def __init__(self, lazy_diff: List[Dict[str, str]], result, stack_items_added: int = 0) -> None:
        super().__init__(stack_items_added)
        self.lazy_diff = lazy_diff
        self.result = result

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        from pytezos.michelson.program import MichelsonProgram

        stack.clear()
        entrypoint, parameter_literal, storage_literal = cls.args  # type: ignore
        entrypoint_str = entrypoint.get_string()  # type: ignore

        parameter = parameter_literal.as_micheline_expr()
        storage = storage_literal.as_micheline_expr()

        program = MichelsonProgram.load(context, with_code=True).instantiate(entrypoint_str, parameter, storage)  # type: ignore
        program.begin(stack, stdout, context)
        program.execute(stack, stdout, context)  # type: ignore
        operations, storage, lazy_diff, res = program.end(stack, stdout)

        return cls(lazy_diff=lazy_diff, result=res)


class PatchInstruction(MichelsonInstruction, prim='PATCH', args_len=1):

    allowed_primitives = ['AMOUNT', 'BALANCE', 'CHAIN_ID', 'SENDER', 'SOURCE', 'NOW']

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):

        res_type: MichelsonType
        res_type = cls.args[0]  # type: ignore

        if res_type.prim == 'AMOUNT':
            context.amount = None  # type: ignore
        elif res_type.prim == 'BALANCE':
            context.balance = None  # type: ignore
        elif res_type.prim == 'CHAIN_ID':
            context.chain_id = None  # type: ignore
        elif res_type.prim == 'SENDER':
            context.sender = None  # type: ignore
        elif res_type.prim == 'SOURCE':
            context.source = None  # type: ignore
        elif res_type.prim == 'NOW':
            context.now = None  # type: ignore
        else:
            raise ValueError(f'Expected one of {cls.allowed_primitives}, got {res_type.prim}')
        return cls()


class PatchValueInstruction(MichelsonInstruction, prim='PATCH', args_len=2):

    allowed_primitives = ['AMOUNT', 'BALANCE', 'CHAIN_ID', 'SENDER', 'SOURCE', 'NOW']

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res_type: MichelsonType
        literal: Type[MichelineLiteral]
        res_type, literal = cls.args  # type: ignore

        if res_type.prim == 'AMOUNT':
            context.amount = literal.get_int()  # type: ignore
        elif res_type.prim == 'BALANCE':
            context.balance = literal.get_int()  # type: ignore
        elif res_type.prim == 'CHAIN_ID':
            context.chain_id = literal.get_string()  # type: ignore
        elif res_type.prim == 'SENDER':
            context.sender = literal.get_string()  # type: ignore
        elif res_type.prim == 'SOURCE':
            context.source = literal.get_string()  # type: ignore
        elif res_type.prim == 'NOW':
            try:
                context.now = literal.get_int()  # type: ignore
            # FIXME: Why does TypeError appear to be wrapped?
            except (TypeError, MichelsonRuntimeError):
                context.now = int(strict_rfc3339.rfc3339_to_timestamp(literal.get_string()))  # type: ignore
        else:
            raise ValueError(f'Expected one of {cls.allowed_primitives}, got {res_type.prim}')
        return cls()


class ResetInstruction(MichelsonInstruction, prim='RESET'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        context.shell = None  # type: ignore
        context.network = None  # type: ignore
        context.chain_id = None  # type: ignore
        context.big_maps = {}  # type: ignore
        stack.items = []
        return cls()


class ResetValueInstruction(MichelsonInstruction, prim='RESET', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        literal: Type[MichelineLiteral]
        literal = cls.args[0]  # type: ignore

        shell = literal.get_string()
        if shell not in nodes:
            raise Exception(f'Expected one of {nodes}, got {shell}')

        if shell.endswith('.pool'):
            shell = shell.split('.')[0]
            assert shell in nodes, f'unknown network {shell}'
            context.shell = ShellQuery(RpcMultiNode(nodes[shell]))  # type: ignore
        elif shell in nodes:
            context.shell = ShellQuery(RpcNode(nodes[shell][0]))  # type: ignore
        else:
            context.shell = ShellQuery(RpcNode(shell))  # type: ignore

        context.network = shell  # type: ignore
        context.chain_id = context.shell.chains.main.chain_id()  # type: ignore
        context.big_maps = {}  # type: ignore
        stack.items = []
        return cls()


class BigMapDiffInstruction(MichelsonInstruction, prim='BIG_MAP_DIFF'):
    def __init__(self, lazy_diff: List[Dict[str, str]], stack_items_added: int = 0) -> None:
        super().__init__(stack_items_added)
        self.lazy_diff = lazy_diff

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        lazy_diff = []  # type: ignore
        # FIXME: AssertionError instead of informational exception
        with suppress(AssertionError):
            stack.peek().aggregate_lazy_diff(lazy_diff)
        stdout.append(f'BIG_MAP_DIFF')
        return cls(lazy_diff=lazy_diff, stack_items_added=1)



from contextlib import suppress
from functools import wraps
from pprint import pformat
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union
from typing import cast
from typing import overload

from pytezos.michelson.forge import unforge_address
from pytezos.michelson.forge import unforge_chain_id
from pytezos.michelson.forge import unforge_micheline
from pytezos.michelson.forge import unforge_public_key
from pytezos.michelson.forge import unforge_signature
from pytezos.michelson.format import micheline_to_michelson


class MichelsonRuntimeError(Exception):
    def format_stdout(self):
        offset, instruction = next(
            ((len(self.args) - i, arg) for i, arg in enumerate(reversed(self.args)) if str(arg).isupper()), (0, 'ERROR')
        )
        message = ' -> '.join(self.args[offset:])
        return f'{instruction}: {message}'


def catch(prim, func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if not e.args:
                e.args = (type(e).__name__,)
            if prim:
                e.args = (prim, *e.args)
            raise MichelsonRuntimeError(*e.args) from e

    return wrapper


def try_catch(prim):
    def _catch(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise MichelsonRuntimeError(prim, *e.args) from e

        return wrapper

    return _catch


class ErrorTrace(type):
    def __new__(mcs, name, bases, attrs, **kwargs):
        wrapped_attrs = {}
        for attr_name, attr in attrs.items():
            prim = kwargs.get('prim')
            if type(attr) in [classmethod, staticmethod] and not attr_name.startswith('_'):
                attr = type(attr)(catch(prim, attr.__func__))
            elif callable(attr) and not attr_name.startswith('_'):
                attr = catch(prim, attr)
            wrapped_attrs[attr_name] = attr
        return type.__new__(mcs, name, bases, wrapped_attrs, **kwargs)


def parse_micheline_prim(prim_expr) -> Tuple[str, list, list]:
    assert isinstance(prim_expr, dict), f'expected dict, got {pformat(prim_expr)} (instr_expr)'
    prim = prim_expr.get('prim')
    assert prim is not None, f'prim field is absent'
    args = prim_expr.get('args', [])
    assert isinstance(args, list), f'{prim}: expected list of args, got {pformat(args)} (args)'
    annots = prim_expr.get('annots', [])
    assert isinstance(annots, list), f'{prim}: expected list of annots, got {pformat(annots)} (annots)'
    return prim, args, annots


def parse_micheline_value(val_expr, handlers: Dict[Tuple[str, int], Callable]):
    assert isinstance(val_expr, dict), f'expected dict, got {pformat(val_expr)} (val_expr)'
    prim, args = val_expr.get('prim'), val_expr.get('args', [])
    expected = ' or '.join(map(lambda x: f'{x[0]} ({x[1]} args)', handlers))
    assert (prim, len(args)) in handlers, f'expected {expected}, got {prim} ({len(args)} args)'
    handler = handlers[(prim, len(args))]  # type: ignore
    return handler(args)


def parse_micheline_literal(val_expr, handlers: Dict[str, Callable]):
    assert isinstance(val_expr, dict), f'expected dict, got {pformat(val_expr)}'
    try:
        core_type, value = next((k, v) for k, v in val_expr.items() if k[0] != '_' and k != 'annots')
    except StopIteration as e:
        raise Exception(f'Can\'t parse literal `{val_expr}`') from e
    expected = ' or '.join(map(lambda x: f'`{x}`', handlers))
    if core_type not in handlers:
        raise Exception(f'Expected one of {expected}, got {core_type}')
    handler = handlers[core_type]
    return handler(value)


def micheline_value_to_python_object(val_expr):
    if isinstance(val_expr, dict):
        if len(val_expr) == 1:
            prim = next(iter(val_expr))
            if prim == 'string':
                return val_expr[prim]
            elif prim == 'int':
                return int(val_expr[prim])
            elif prim == 'bytes':
                return blind_unpack(bytes.fromhex(val_expr[prim]))
            elif prim == 'bool':
                return val_expr[prim] == 'True'
        elif val_expr.get('prim'):
            prim = val_expr['prim']
            if prim == 'Pair':
                return tuple(micheline_value_to_python_object(x) for x in val_expr['args'])
    return micheline_to_michelson(val_expr)


def blind_unpack(data: bytes):
    with suppress(ValueError):
        return unforge_chain_id(data)
    with suppress(ValueError, KeyError):
        return unforge_address(data)
    with suppress(ValueError, KeyError):
        return unforge_public_key(data)
    with suppress(ValueError):
        return unforge_signature(data)

    if len(data) > 0 and data.startswith(b'\x05'):
        with suppress(ValueError, AssertionError):
            res = unforge_micheline(data[1:])
            return micheline_value_to_python_object(res)

    with suppress(ValueError):
        return data.decode()

    return data


class Micheline(metaclass=ErrorTrace):
    prim: Optional[str] = None
    args: List[Type['Micheline']] = []
    literal: Optional[Union[int, str, bytes]] = None
    classes: Dict[Tuple[str, Optional[int]], Type['Micheline']] = {}

    @classmethod
    def __init_subclass__(cls, prim: Optional[str] = None, args_len: Optional[int] = 0, **kwargs):
        super().__init_subclass__(**kwargs)  # type: ignore
        if prim is not None:
            assert (prim, args_len) not in cls.classes, f'duplicate key {prim} ({args_len} args)'
            cls.classes[(prim, args_len)] = cls
            cls.prim = prim

    def __str__(self):
        raise AssertionError('__str__ has to be explicitly defined')

    def __repr__(self):
        raise AssertionError('__repr__ has to be explicitly defined')

    @staticmethod
    def match(expr) -> Type['Micheline']:
        if isinstance(expr, list):
            args = [Micheline.match(arg) for arg in expr]
            return MichelineSequence.create_type(args=args)
        elif isinstance(expr, dict):
            if expr.get('prim'):
                prim, args, annots = parse_micheline_prim(expr)

                # FIXME: We need entrypoint to be an argument
                if prim == 'RUN':
                    if annots:
                        args = [{'string': annots[0][1:]}] + args  # type: ignore
                        annots = []
                    else:
                        args = [{'string': 'default'}] + args  # type: ignore

                args_len = len(args)
                if (prim, args_len) not in Micheline.classes:
                    args_len = None  # type: ignore
                assert (prim, args_len) in Micheline.classes, f'unregistered primitive {prim} ({args_len} args)'
                cls = Micheline.classes[prim, args_len]
                try:
                    return cls.create_type(args=list(map(Micheline.match, args)), annots=annots)
                except Exception as e:
                    raise MichelsonRuntimeError(cls.prim, *e.args) from e
            else:
                literal = parse_micheline_literal(
                    expr,
                    {
                        'int': int,
                        'string': str,
                        'bytes': bytes.fromhex,
                    },
                )
                return MichelineLiteral.create(literal=literal)
        else:
            raise MichelsonRuntimeError(f'malformed expression `{expr}`')

    @classmethod
    def create_type(
        cls,
        args: List[Type['Micheline']],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['Micheline']:
        res = type(cls.__name__, (cls,), dict(args=args, **kwargs))
        return cast(Type['MichelsonPrimitive'], res)  # type: ignore

    @classmethod
    def assert_type_equal(cls, other: Type['Micheline'], path='', message=''):
        comment = f' [{message}]' if message else ''
        assert cls.prim == other.prim, f'expected {other.prim}, got {cls.prim} at `{path}`{comment}'
        assert len(cls.args) == len(
            other.args
        ), f'expected {len(other.args)} args, got {len(cls.args)} at `{path}`{comment}'
        for i, arg in enumerate(other.args):
            cls.args[i].assert_type_equal(arg, path=f'{path}/{i}', message=message)
        assert (
            cls.literal == other.literal
        ), f'expected {other.literal}, got {cls.literal} at `{path}`{comment}'  # type: ignore

    @classmethod
    def assert_type_in(cls, *others: Type['Micheline'], message=''):
        comment = f' [{message}]' if message else ''
        expected = [ty.prim for ty in others]
        assert any(issubclass(cls, ty) for ty in others), f'expected one of {expected}, got {cls.prim}{comment}'

    @classmethod
    def as_micheline_expr(cls) -> dict:
        args = [arg.as_micheline_expr() for arg in cls.args]
        expr = {'prim': cls.prim, 'args': args}
        return {k: v for k, v in expr.items() if v}

    @classmethod
    def execute(cls, stack, stdout, context) -> 'Micheline':
        raise AssertionError(f'`execute` has to be explicitly defined ({cls.prim})')


MichelineT = TypeVar('MichelineT', bound=Micheline)


class MichelineSequence(Micheline):
    def __init__(self, items: List[Micheline]):
        super(MichelineSequence, self).__init__()
        self.items = items

    @classmethod
    def as_micheline_expr(cls) -> list:  # type: ignore
        return [arg.as_micheline_expr() for arg in cls.args]

    @classmethod
    def execute(cls, stack, stdout, context) -> Micheline:
        return cls([arg.execute(stack, stdout, context) for arg in cls.args])


class GlobalConstant(Micheline, prim='constant', args_len=1):
    @classmethod
    def create_type(
        cls,
        args: List[Type['Micheline']],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['Micheline']:
        raise RuntimeError('Please, register global constants in advance using context helpers')


class MichelineLiteral(Micheline):
    @classmethod
    def create(cls, literal: Union[int, str, bytes]):
        return cls.create_type(args=[], annots=[], literal=literal)

    @classmethod
    def as_micheline_expr(cls) -> dict:
        if isinstance(cls.literal, int):
            return {'int': str(cls.literal)}
        elif isinstance(cls.literal, str):
            return {'string': cls.literal}
        elif isinstance(cls.literal, bytes):
            return {'bytes': cls.literal.hex()}
        else:
            raise AssertionError(f'unexpected value `{cls.literal}`')

    @classmethod
    def get_int(cls) -> int:
        if not isinstance(cls.literal, int):
            raise TypeError(f'Expected int, got {cls.literal}')  # type: ignore
        return cls.literal

    @classmethod
    def get_string(cls) -> str:
        if not isinstance(cls.literal, str):
            raise TypeError(f'Expected str, got {cls.literal}')  # type: ignore
        return cls.literal

    @classmethod
    def get_bytes(cls) -> bytes:
        if not isinstance(cls.literal, bytes):
            raise TypeError(f'Expected bytes, got {cls.literal}')  # type: ignore
        return cls.literal


def validate_sections(sequence: Type[MichelineSequence], required_sections: Sequence[str]) -> None:
    if len(sequence.args) < len(required_sections):
        raise Exception(f'expected at least {len(required_sections)} sections, got {len(sequence.args)}')
    sequence_sections = {arg.prim for arg in sequence.args}
    if set(required_sections).intersection(sequence_sections) == required_sections:
        raise Exception(f'Unknown sections {sequence_sections}, required: {required_sections}')


@overload
def get_script_section(
    sequence: Type[MichelineSequence],
    cls: None,
    name: str,
    required: Literal[True],
) -> MichelineT:
    ...


@overload
def get_script_section(
    sequence: Type[MichelineSequence],
    cls: None,
    name: str,
    required: Literal[False],
) -> Optional[MichelineT]:
    ...


@overload
def get_script_section(
    sequence: Type[MichelineSequence],
    cls: Type[MichelineT],
    name: None,
    required: Literal[True],
) -> MichelineT:
    ...


@overload
def get_script_section(
    sequence: Type[MichelineSequence],
    cls: Type[MichelineT],
    name: None,
    required: Literal[False],
) -> Optional[MichelineT]:
    ...


def get_script_section(
    sequence,
    cls=None,
    name=None,
    required=False,
):
    try:
        if cls:
            return next(arg for arg in sequence.args if issubclass(arg, cls))
        elif name:
            return next(arg for arg in sequence['code'] if arg['prim'] == name)
        else:
            raise Exception('Either `cls` or `name` must be specified')
    except StopIteration:
        if required:
            raise
    return None


@overload
def get_script_sections(
    sequence: Type[MichelineSequence],
    cls: None,
    name: str,
) -> List[MichelineT]:
    ...


@overload
def get_script_sections(
    sequence: Type[MichelineSequence],
    cls: Type[MichelineT],
    name: None,
) -> List[MichelineT]:
    ...


def get_script_sections(sequence, cls=None, name=None):
    if cls:
        return [arg for arg in sequence.args if issubclass(arg, cls)]
    elif name:
        return [arg for arg in sequence['code'] if arg['prim'] == name]
    else:
        raise Exception('Either `cls` or `name` must be specified')



from typing import Any
from typing import Dict
from typing import List

from pytezos.context.impl import ExecutionContext
from pytezos.michelson.program import MichelsonProgram
from pytezos.operation.result import OperationResult


class ContractCallResult(OperationResult):
    """Encapsulates the result of a contract invocation."""

    @classmethod
    def from_run_operation(
        cls,
        operation_group: Dict[str, Any],
        context: ExecutionContext,
    ) -> List['ContractCallResult']:
        """Get a list of results from an operation group content with metadata.

        :param operation_group: {..., "contents": [{..., kind: "transaction", ...}]}
        :param context: execution context
        :rtype: ContractCallResult
        """
        results: List['OperationResult'] = []
        for content in OperationResult.iter_contents(operation_group):
            if content['kind'] == 'transaction':
                if content['destination'] == context.address:
                    results.append(cls.from_transaction(content))
            elif content['kind'] == 'origination':
                result = cls.get_result(content)
                if context.address in result.get('originated_contracts', []):
                    results.append(cls.from_origination(content))

        program = MichelsonProgram.load(context)

        def decode_result(res: OperationResult) -> 'ContractCallResult':
            kwargs = {}  # type: ignore
            if hasattr(res, 'storage') and res.storage is not None:  # type: ignore
                storage = program.storage.from_micheline_value(res.storage)  # type: ignore
                if hasattr(res, 'lazy_storage_diff'):
                    kwargs.update(lazy_diff=res.lazy_diff)  # type: ignore
                    storage = storage.merge_lazy_diff(res.lazy_diff)  # type: ignore
                kwargs.update(storage=storage.to_python_object())
            if hasattr(res, 'parameters'):
                parameters = program.parameter.from_parameters(res.parameters)  # type: ignore
                kwargs.update(parameters=parameters)
            if hasattr(res, 'operations'):
                kwargs.update(operations=res.operations)  # type: ignore
            return cls(**kwargs)

        return list(map(decode_result, results))

    @classmethod
    def from_run_code(cls, response: Dict[str, Any], parameters, context: ExecutionContext) -> 'ContractCallResult':
        """Parse a result of :py:meth:`pytezos.contract.call.ContractCall.run_code` execution.

        :param response: RPC response (JSON)
        :param parameters: {"entrypoint": str, "value": $Micheline}
        :param context: execution context
        :rtype: ContractCallResult
        """
        program = MichelsonProgram.load(context)
        parameters = program.parameter.from_parameters(parameters)
        storage = program.storage.from_micheline_value(response['storage'])
        extended_storage = storage.merge_lazy_diff(response.get('lazy_storage_diff', []))
        return cls(
            parameters=parameters.to_python_object(),
            storage=extended_storage.to_python_object(lazy_diff=True),
            lazy_diff=response.get('lazy_storage_diff', []),
            operations=response.get('operations', []),
        )



from contextlib import suppress
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Type

import dateutil.parser
from cattr import Converter


def structure_datetime(obj: Any, cls: Type) -> datetime:
    with suppress(ValueError, TypeError):
        return datetime.utcfromtimestamp(float(obj)).replace(tzinfo=timezone.utc)
    with suppress(dateutil.parser.ParserError):
        return dateutil.parser.parse(obj)
    raise ValueError


def unstructure_datetime(obj: datetime) -> float:
    return obj.timestamp()


converter = Converter()
converter.register_structure_hook(datetime, structure_datetime)
converter.register_unstructure_hook(datetime, unstructure_datetime)



from os.path import exists
from os.path import expanduser
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from pytezos.context.impl import ExecutionContext
from pytezos.crypto.encoding import is_pkh
from pytezos.crypto.encoding import is_public_key
from pytezos.crypto.key import Key
from pytezos.crypto.key import is_installed
from pytezos.jupyter import InlineDocstring
from pytezos.rpc import RpcMultiNode
from pytezos.rpc import RpcNode
from pytezos.rpc import ShellQuery
from pytezos.rpc.errors import RpcError

# NOTE: Built-in key for PyTezos client, please, use responsibly.
default_network = 'ghostnet'
default_key = 'edsk33N474hxzA4sKeWVM6iuGNGDpX2mGwHNxEA4UbWS8sW3Ta3NKH'
default_key_hash = 'tz1grSQDByRpnVs7sPtaprNZRp531ZKz6Jmm'

# NOTE: For flextesa sandbox
alice_key = 'edsk3QoqBuvdamxouPhin7swCvkQNgq4jP5KZPbwWNnwdZpSpJiEbq'
alice_key_hash = 'tz1VSUr8wwNhLAzempoch5d6hLRiTh8Cjcjb'

# NOTE: For tezos-node in sandboxed mode
dictator_key = 'edsk31vznjHSSpGExDMHYASz45VZqXN4DPxvsa4hAyY8dHM28cZzp6'

nodes = {
    'mainnet': ['https://rpc.tzkt.io/mainnet'],
    'sandbox': ['http://127.0.0.1:8732/'],
    'sandboxnet': ['http://127.0.0.1:8732/'],
    'localhost': ['http://127.0.0.1:8732/'],
    'ghostnet': ['https://rpc.tzkt.io/ghostnet'],
    'jakartanet': ['https://rpc.tzkt.io/jakartanet'],
    'kathmandunet': ['https://rpc.tzkt.io/kathmandunet'],
}
keys = {
    'alice': alice_key,
    'dictator': dictator_key,
    'activator': dictator_key,
    'bootstrap1': 'edsk3gUfUPyBSfrS9CCgmCiQsTCHGkviBDusMxDJstFtojtc1zcpsh',
    'bootstrap2': 'edsk39qAm1fiMjgmPkw1EgQYkMzkJezLNewd7PLNHTkr6w9XA2zdfo',
    'bootstrap3': 'edsk4ArLQgBTLWG5FJmnGnT689VKoqhXwmDPBuGx3z4cvwU9MmrPZZ',
    'bootstrap4': 'edsk2uqQB9AY4FvioK2YMdfmyMrer5R8mGFyuaLLFfSRo8EoyNdht3',
    'bootstrap5': 'edsk4QLrcijEffxV31gGdN2HU7UpyJjA8drFoNcmnB28n89YjPNRFm',
}


class KeyHash(Key):
    def __init__(self, public_key_hash):
        super().__init__(b'\x00' * 32)
        self._pkh = public_key_hash

    def __repr__(self):
        res = [
            super().__repr__(),
            f'\nPublic key hash',
            self.public_key_hash(),
        ]
        return '\n'.join(res)

    def public_key_hash(self):
        return self._pkh

    def public_key(self):
        raise NotImplementedError("Use private key instead of a public key hash")

    def secret_key(self, passphrase=None, ed25519_seed=True):
        raise NotImplementedError("Use private key instead of a public key hash")

    def sign(self, message, generic=False):
        raise NotImplementedError("Use private key instead of a public key hash")

    def verify(self, signature, message):
        raise NotImplementedError("Use private key instead of a public key hash")


class ContextMixin(metaclass=InlineDocstring):
    """Mixin for blockchain interaction, stores node connection and key object."""

    def __init__(self, context: Optional[ExecutionContext] = None):
        super().__init__()
        if context is None:
            context = ExecutionContext(
                shell=ShellQuery(RpcNode(nodes[default_network][0])),
                key=Key.from_encoded_key(default_key) if is_installed() else KeyHash(default_key_hash),
            )
        self.context = context

    @property
    def shell(self) -> ShellQuery:
        assert self.context.shell, f'network is undefined'
        return self.context.shell

    @property
    def key(self) -> Key:
        assert self.context.key, f'key is undefined'
        return self.context.key

    @property
    def address(self) -> Optional[str]:
        return self.context.address

    @property
    def block_id(self) -> Union[str, int]:
        return self.context.block_id

    def __repr__(self):
        res = [super().__repr__(), '\nProperties']
        if self.context.key is not None:
            res.append(f'.key\t\t{self.key.public_key_hash()}')
        if self.context.shell is not None:
            res.append(f'.shell\t\t{self.shell.node.uri}')
        if self.context.address is not None:
            res.append(f'.address\t{self.address}')
        if self.context.block_id is not None:
            res.append(f'.block_id\t{self.block_id}')
        return '\n'.join(res)

    def _spawn_context(
        self,
        shell: Optional[Union[ShellQuery, str]] = None,
        key: Optional[Union[Key, str, dict]] = None,
        address: Optional[str] = None,
        block_id: Optional[Union[str, int]] = None,
        mode: Optional[str] = None,
        script: Optional[dict] = None,
        ipfs_gateway: Optional[str] = None,
        balance: Optional[int] = None,
        view_results: Optional[Dict[str, Any]] = None,
    ) -> ExecutionContext:
        if isinstance(shell, str):
            if shell.endswith('.pool'):
                shell = shell.split('.')[0]
                assert shell in nodes, f'unknown network {shell}'
                shell = ShellQuery(RpcMultiNode(nodes[shell]))
            elif shell in nodes:
                shell = ShellQuery(RpcNode(nodes[shell][0]))
            else:
                shell = ShellQuery(RpcNode(shell))
        else:
            assert shell is None or isinstance(shell, ShellQuery), f'unexpected shell {shell}'

        if isinstance(key, str):
            if key in keys:
                key = Key.from_encoded_key(keys[key])
            elif is_public_key(key):
                key = Key.from_encoded_key(key)
            elif is_pkh(key):
                key = KeyHash(key)
            elif exists(expanduser(key)):
                key = Key.from_faucet(key)
            else:
                key = Key.from_alias(key)
        elif isinstance(key, dict):
            key = Key.from_faucet(key)
        else:
            assert key is None or isinstance(key, Key), f'unexpected key {key}'

        if script is None and isinstance(address, str):
            try:
                script = self.shell.contracts[address].script()
            except RpcError as e:
                raise RpcError(f'Contract {address} not found', *e.args) from e

        return ExecutionContext(
            shell=shell or self.context.shell,
            key=key or self.context.key,
            address=address,
            block_id=block_id,
            script=script or self.context.script,
            mode=mode or self.context.mode,
            ipfs_gateway=ipfs_gateway,
            balance=balance or self.context.balance,
            view_results=view_results,
        )



from decimal import Decimal
from pprint import pformat
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from deprecation import deprecated  # type: ignore

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.contract.result import ContractCallResult
from pytezos.jupyter import get_class_docstring
from pytezos.logging import logger
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.repl import Interpreter
from pytezos.michelson.sections.storage import StorageSection
from pytezos.operation import DEFAULT_BURN_RESERVE
from pytezos.operation import DEFAULT_GAS_RESERVE
from pytezos.operation.content import format_mutez
from pytezos.operation.content import format_tez
from pytezos.operation.group import OperationGroup


def skip_nones(**kwargs) -> dict:
    return {k: v for k, v in kwargs.items() if v is not None}


class ContractCall(ContextMixin):
    """Proxy class encapsulating a contract call: contract type scheme, contract address, parameters, and amount"""

    def __init__(self, context: ExecutionContext, parameters: dict, amount: Union[int, Decimal] = 0) -> None:
        super().__init__(context=context)
        self.parameters = parameters
        self.amount = amount

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            f'.amount\t{self.amount}',
            '\nParameters',
            pformat(self.parameters),
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def with_amount(self, amount: Union[int, Decimal]) -> 'ContractCall':
        """Set amount of funds to send with transaction to the contract.

        :param amount: amount in microtez (int) or tez (Decimal)
        :rtype: ContractCall
        """
        return ContractCall(
            context=self.context,
            parameters=self.parameters,
            amount=amount,
        )

    def as_transaction(self, **kwargs) -> OperationGroup:
        """Get operation content.

        :rtype: OperationGroup
        """
        return OperationGroup(context=self._spawn_context()).transaction(
            destination=self.address, amount=self.amount, parameters=self.parameters, **kwargs
        )

    @property  # type: ignore
    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `as_transaction()` instead')
    def operation_group(self) -> OperationGroup:
        return self.as_transaction().fill()

    def send(
        self,
        gas_reserve: int = DEFAULT_GAS_RESERVE,
        burn_reserve: int = DEFAULT_BURN_RESERVE,
        min_confirmations: int = 0,
        ttl: Optional[int] = None,
    ) -> 'OperationGroup':
        """Fill, sign, and broadcast the transaction

        :param gas_reserve: Add a safe reserve for dynamically calculated gas limit (default is 100).
        :param burn_reserve: Add a safe reserve for dynamically calculated storage limit (default is 100).
        :param min_confirmations: number of block injections to wait for before returning (default is 0, i.e. async mode)
        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :return: OperationGroup with hash filled
        """
        return self.as_transaction().send(
            gas_reserve=gas_reserve,
            burn_reserve=burn_reserve,
            min_confirmations=min_confirmations,
            ttl=ttl,
        )

    def send_async(
        self,
        ttl: int,
        counter: int,
        gas_limit: int,
        storage_limit: int,
        minimal_nanotez_per_gas_unit: Optional[int] = None,
    ) -> 'OperationGroup':
        """
        Send operation without simulation or pre-validation

        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :param counter: Set counter value
        :param gas_limit: Set gas_limit value
        :param storage_limit: Set storage_limit value
        :param minimal_nanotez_per_gas_unit: Override minimal_nanotez_per_gas_unit constant
        :rtype: OperationGroup
        """
        return self.as_transaction().send_async(
            ttl=ttl,
            counter=counter,
            gas_limit=gas_limit,
            storage_limit=storage_limit,
            minimal_nanotez_per_gas_unit=minimal_nanotez_per_gas_unit,
        )

    @deprecated(deprecated_in='3.2.2', removed_in='4.0.0', details='use `send()` instead')
    def inject(self, _async=True, preapply=True, check_result=True, num_blocks_wait=5) -> OperationGroup:
        """Send operation to blockchain."""
        return (
            self.as_transaction()
            .autofill()
            .sign()
            .inject(
                _async=_async,
                preapply=preapply,
                check_result=check_result,
                num_blocks_wait=num_blocks_wait,
            )
        )

    def cmdline(self) -> str:
        """Generate command line for tezos-client."""
        arg = micheline_to_michelson(self.parameters['value'], inline=True)
        source = self.key.public_key_hash()
        amount = format_tez(self.amount)
        entrypoint = self.parameters['entrypoint']
        return f'transfer {amount} from {source} to {self.address} --entrypoint \'{entrypoint}\' --arg \'{arg}\''

    def interpret(
        self,
        storage=None,
        source=None,
        sender=None,
        amount=None,
        balance=None,
        chain_id=None,
        level=None,
        now=None,
        self_address=None,
        view_results: Optional[Dict[str, Any]] = None,
    ) -> ContractCallResult:
        """Run code in the builtin REPL (WARNING! Not recommended for critical tasks).

        :param storage: initial storage as Python object, leave None if you want to generate a dummy one
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param amount: patch AMOUNT
        :param balance: patch BALANCE
        :param chain_id: patch CHAIN_ID
        :param level: patch LEVEL
        :param now: patch NOW
        :param self_address: patch SELF/SELF_ADDRESS
        :param view_results: patch VIEW calls (keys must be string "address%view", values => Python objects)
        :rtype: pytezos.contract.result.ContractCallResult
        """
        storage_ty = StorageSection.match(self.context.storage_expr)
        if storage is None:
            initial_storage = storage_ty.dummy(self.context).to_micheline_value(lazy_diff=True)
        else:
            initial_storage = storage_ty.from_python_object(storage).to_micheline_value(lazy_diff=True)
        assert self.context.script
        operations, storage, lazy_diff, stdout, error = Interpreter.run_code(
            parameter=self.parameters['value'],
            entrypoint=self.parameters['entrypoint'],
            storage=initial_storage,
            script=self.context.script['code'],
            source=source,
            sender=sender or source,
            amount=amount or self.amount,
            balance=balance,
            chain_id=chain_id,
            level=level,
            now=now,
            address=self_address,
            view_results=view_results,
        )
        if error:
            logger.debug('\n'.join(stdout))
            raise error
        res = {
            'operations': operations,
            'storage': storage,
            'lazy_storage_diff': lazy_diff,
        }
        return ContractCallResult.from_run_code(
            res,
            parameters=self.parameters,
            context=self.context,
        )

    def run_code(
        self,
        storage=None,
        source=None,
        sender=None,
        amount=None,
        balance=None,
        chain_id=None,
        gas_limit=None,
    ) -> ContractCallResult:
        """Execute using RPC interpreter.

        :param storage: initial storage as Python object, leave None if you want to generate a dummy one
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param amount: patch AMOUNT
        :param balance: patch BALANCE
        :param chain_id: patch CHAIN_ID
        :param gas_limit: restrict max consumed gas
        :rtype: ContractCallResult
        """
        storage_ty = StorageSection.match(self.context.storage_expr)
        if storage is None:
            initial_storage = storage_ty.dummy(self.context).to_micheline_value(lazy_diff=True)
        else:
            initial_storage = storage_ty.from_python_object(storage).to_micheline_value(lazy_diff=True)
        script = [self.context.parameter_expr, self.context.storage_expr, self.context.code_expr]
        query = skip_nones(
            script=script,
            storage=initial_storage,
            entrypoint=self.parameters['entrypoint'],
            input=self.parameters['value'],
            amount=format_mutez(amount or self.amount),
            chain_id=chain_id or self.context.get_chain_id(),
            source=sender,
            payer=source,
            balance=str(balance or 0),
            gas=str(gas_limit) if gas_limit is not None else None,
        )
        res = self.shell.blocks[self.block_id].helpers.scripts.run_code.post(query)
        return ContractCallResult.from_run_code(res, parameters=self.parameters, context=self.context)

    def run_operation(self) -> ContractCallResult:
        """Simulate operation using real context.

        :rtype: ContractCallResult
        """
        opg_with_metadata = self.as_transaction().fill().run()
        results = ContractCallResult.from_run_operation(opg_with_metadata, context=self.context)
        assert len(results) == 1
        return results[0]

    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use either `run_code` or `run_operation`')
    def result(self, storage=None, source=None, sender=None, gas_limit=None) -> ContractCallResult:
        """Simulate operation and parse the result.

        :param storage: Python object only. If storage is specified, `run_code` is called instead of `run_operation`.
        :param source: Can be specified for unit testing purposes
        :param sender: Can be specified for unit testing purposes, \
        see https://tezos.gitlab.io/whitedoc/michelson.html#operations-on-contracts for the difference
        :param gas_limit: Specify gas limit (default is gas hard limit)
        :rtype: ContractCallResult
        """
        if storage or source or sender or gas_limit:
            return self.run_code(storage=storage, source=source, sender=sender, gas_limit=gas_limit)
        return self.run_operation()

    def callback_view(self, storage=None):
        """Get return value of an on-chain callback method.

        :param storage: initial storage as Python object,
                        if None then the current one will be taken (if context attached), otherwise empty (dummy)
        :returns: Decoded parameters of a callback
        """
        if storage:
            storage_ty = StorageSection.match(self.context.storage_expr)
            initial_storage = storage_ty.from_python_object(storage).to_micheline_value(lazy_diff=True)
        elif self.address:
            initial_storage = self.shell.blocks[self.context.block_id].context.contracts[self.address].storage()
        else:
            storage_ty = StorageSection.match(self.context.storage_expr)
            initial_storage = storage_ty.dummy(self.context).to_micheline_value(lazy_diff=True)

        operations, _, stdout, error = Interpreter.run_callback(
            parameter=self.parameters['value'],
            entrypoint=self.parameters['entrypoint'],
            storage=initial_storage,
            context=self.context,
        )
        if error:
            logger.debug('\n'.join(stdout))
            raise error
        if not operations:
            raise Exception('No operation could be interpreted')
        if len(operations) > 1:
            raise Exception('Multiple internal operations, not sure which one to pick')
        return operations[0]

    @deprecated(deprecated_in='3.0.4', removed_in='4.0.0', details='Use callback_view instead')
    def view(self):
        return self.callback_view()



from typing import List
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline


class CodeSection(Micheline, prim='code', args_len=1):
    @staticmethod
    def match(code_expr) -> Type['CodeSection']:
        cls = Micheline.match(code_expr)
        if not issubclass(cls, CodeSection):
            cls = CodeSection.create_type(args=[cls])
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_code_expr(cls.as_micheline_expr())
        stdout.append(f'code: updated')



from typing import List
from typing import Tuple
from typing import Type
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import Wildcard
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types.base import MichelsonType


class PushInstruction(MichelsonInstruction, prim='PUSH', args_len=2):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res_type, literal = cast(Tuple[Type[MichelsonType], Type[Micheline]], cls.args)
        assert res_type.is_pushable(), f'{res_type.prim} contains non-pushable arguments'
        res = res_type.from_literal(literal)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class DropnInstruction(MichelsonInstruction, prim='DROP', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        count = cls.args[0].get_int()  # type: ignore
        dropped = stack.pop(count=count)
        stdout.append(format_stdout(cls.prim, dropped, [], count))  # type: ignore
        return cls()


class DropInstruction(MichelsonInstruction, prim='DROP'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        dropped = stack.pop1()
        stdout.append(format_stdout(cls.prim, [dropped], []))  # type: ignore
        return cls()


class DupnInstruction(MichelsonInstruction, prim='DUP', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        depth = cls.args[0].get_int() - 1  # type: ignore
        stack.protect(count=depth)
        res = stack.peek().duplicate()
        stack.restore(count=depth)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [*Wildcard.n(depth), res], [res, *Wildcard.n(depth), res], depth))  # type: ignore
        return cls(stack_items_added=1)


class DupInstruction(MichelsonInstruction, prim='DUP'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = stack.peek().duplicate()
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [res], [res, res]))  # type: ignore
        return cls(stack_items_added=1)


class SwapInstruction(MichelsonInstruction, prim='SWAP'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = stack.pop2()
        stack.push(a)
        stack.push(b)
        stdout.append(format_stdout(cls.prim, [a, b], [b, a]))  # type: ignore
        return cls(stack_items_added=2)


class DigInstruction(MichelsonInstruction, prim='DIG', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        depth = cls.args[0].get_int()  # type: ignore
        stack.protect(count=depth)
        res = stack.pop1()
        stack.restore(count=depth)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [*Wildcard.n(depth), res], [res, *Wildcard.n(depth)], depth))  # type: ignore
        return cls(stack_items_added=1)


class DugInstruction(MichelsonInstruction, prim='DUG', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        depth = cls.args[0].get_int()  # type: ignore
        res = stack.pop1()
        stack.protect(count=depth)
        stack.push(res)
        stack.restore(count=depth)
        stdout.append(format_stdout(cls.prim, [res, *Wildcard.n(depth)], [*Wildcard.n(depth), res], depth))  # type: ignore
        return cls(stack_items_added=1)


class CastIntruction(MichelsonInstruction, prim='CAST', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = stack.pop1()
        # TODO: will become obsolete in the next protocol? (because annots are no longer part of the type)
        # cast_type = cast(Type[MichelsonType], cls.args[0])
        # res = cast_type.from_micheline_value(top.to_micheline_value())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [res], [res]))  # type: ignore
        return cls(stack_items_added=1)


class RenameInstruction(MichelsonInstruction, prim='RENAME'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        return cls()



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.stack import MichelsonStack


class Wildcard:
    @staticmethod
    def n(count: int) -> List['Wildcard']:
        return [Wildcard() for _ in range(count)]

    def __repr__(self):
        return '*'


def format_stdout(prim: str, inputs: list, outputs: list, arg=None):
    arg = f' {arg}' if arg else ''
    pop = " : ".join(map(repr, inputs)) if inputs else '_'
    push = " : ".join(map(repr, outputs)) if outputs else '_'
    return f'{prim}{arg} / {pop} => {push}'


def dispatch_types(
    *args: Type[Micheline],
    mapping: Dict[Tuple[Type[Micheline], ...], Tuple[Any, ...]],
):
    key = tuple(arg.prim for arg in args)
    mapping = {tuple(arg.prim for arg in k): v for k, v in mapping.items()}  # type: ignore
    assert key in mapping, f'unexpected types `{" * ".join(key)}`'  # type: ignore
    return mapping[key]  # type: ignore


class MichelsonInstruction(Micheline):
    args: List[Union[Type['MichelsonInstruction'], Any]] = []
    field_names: List[str] = []
    var_names: List[str] = []

    def __init__(self, stack_items_added: int = 0) -> None:
        self.stack_items_added = stack_items_added

    @staticmethod
    def match(expr) -> Type['MichelsonInstruction']:
        return cast(Type['MichelsonInstruction'], Micheline.match(expr))

    @classmethod
    def create_type(
        cls,
        args: List[Type['Micheline']],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['MichelsonInstruction']:
        if annots:
            field_names = [a[1:] for a in annots if a.startswith('%')]
            var_names = [a[1:] for a in annots if a.startswith('@')]
        else:
            field_names, var_names = [], []
        res = type(
            cls.__name__,
            (cls,),
            {
                'args': args,
                'field_names': field_names,
                'var_names': var_names,
                **kwargs,
            },
        )
        return cast(Type['MichelsonInstruction'], res)

    @classmethod
    def as_micheline_expr(cls) -> dict:
        annots = []
        if cls.var_names is not None:
            annots.extend([f'@{x}' for x in cls.var_names])
        if cls.field_names is not None:
            annots.extend([f'%{x}' for x in cls.field_names])
        args = [arg.as_micheline_expr() for arg in cls.args]
        expr = {'prim': cls.prim, 'annots': annots, 'args': args}
        return {k: v for k, v in expr.items() if v}

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        raise NotImplementedError



from typing import List
from typing import Optional
from typing import Tuple

from pyblake2 import blake2b  # type: ignore

from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.key import Key


def get_originated_address(index: int, opg_hash=None):
    prefix = base58_decode(opg_hash) if opg_hash else b'\x00' * 32
    nonce = prefix + index.to_bytes(4, 'big')
    nonce_hash = blake2b(data=nonce, digest_size=20).digest()
    return base58_encode(nonce_hash, b'KT1').decode()


class KeyHash(Key):
    def __init__(self, public_key_hash):
        super().__init__(0)
        self._pkh = public_key_hash

    def __repr__(self):
        res = [
            super().__repr__(),
            f'\nPublic key hash',
            self.public_key_hash(),
        ]
        return '\n'.join(res)

    def public_key_hash(self):
        return self._pkh

    def public_key(self):
        raise NotImplementedError

    def secret_key(self, passphrase=None, ed25519_seed=True):
        raise NotImplementedError

    def sign(self, message, generic=False):
        raise NotImplementedError

    def verify(self, signature, message):
        raise NotImplementedError


class AbstractContext:
    def reset(self):
        raise NotImplementedError

    def register_big_map(self, ptr: int, copy=False) -> int:
        raise NotImplementedError

    def get_tmp_big_map_id(self) -> int:
        raise NotImplementedError

    def get_big_map_diff(self, ptr: int) -> Tuple[Optional[int], int, str]:
        raise NotImplementedError

    def get_originated_address(self) -> str:
        raise NotImplementedError

    def spend_balance(self, amount: int):
        raise NotImplementedError

    def get_parameter_expr(self, address=None) -> Optional:  # type: ignore
        raise NotImplementedError

    def get_storage_expr(self, address=None) -> Optional:  # type: ignore
        raise NotImplementedError

    def get_storage_value(self, address=None) -> Optional:  # type: ignore
        raise NotImplementedError

    def get_code_expr(self):
        raise NotImplementedError

    def get_view_result(self, name, address=None) -> Optional:  # type: ignore
        raise NotImplementedError

    def get_view_expr(self, name, address=None) -> Optional:  # type: ignore
        raise NotImplementedError

    def get_views_expr(self) -> List:  # type: ignore
        raise NotImplementedError

    def get_input_expr(self):
        raise NotImplementedError

    def get_output_expr(self):
        raise NotImplementedError

    def get_sender_expr(self):
        raise NotImplementedError

    def get_balance_expr(self):
        raise NotImplementedError

    def get_amount_expr(self):
        raise NotImplementedError

    def get_self_expr(self):
        raise NotImplementedError

    def get_now_expr(self):
        raise NotImplementedError

    def get_source_expr(self):
        raise NotImplementedError

    def get_chain_id_expr(self):
        raise NotImplementedError

    def get_big_maps_expr(self):
        raise NotImplementedError

    def set_storage_expr(self, type_expr):
        raise NotImplementedError

    def set_parameter_expr(self, type_expr):
        raise NotImplementedError

    def set_code_expr(self, code_expr):
        raise NotImplementedError

    def set_view_expr(self, name, view_expr):
        raise NotImplementedError

    def set_input_expr(self, code_expr):
        raise NotImplementedError

    def set_output_expr(self, code_expr):
        raise NotImplementedError

    def set_sender_expr(self, type_expr):
        raise NotImplementedError

    def set_balance_expr(self, type_expr):
        raise NotImplementedError

    def set_amount_expr(self, code_expr):
        raise NotImplementedError

    def set_self_expr(self, code_expr):
        raise NotImplementedError

    def set_now_expr(self, code_expr):
        raise NotImplementedError

    def set_source_expr(self, code_expr):
        raise NotImplementedError

    def set_chain_id_expr(self, code_expr):
        raise NotImplementedError

    def set_big_maps_expr(self, code_expr):
        raise NotImplementedError

    def get_big_map_value(self, ptr: int, key_hash: str):
        raise NotImplementedError

    def register_sapling_state(self, ptr: int):
        raise NotImplementedError

    def get_tmp_sapling_state_id(self) -> int:
        raise NotImplementedError

    def get_sapling_state_diff(self, offset_commitment=0, offset_nullifier=0) -> Tuple[int, list]:
        raise NotImplementedError

    def get_self_address(self) -> str:
        raise NotImplementedError

    def get_amount(self) -> int:
        raise NotImplementedError

    def get_voting_power(self, address: str) -> int:
        raise NotImplementedError

    def get_total_voting_power(self) -> int:
        raise NotImplementedError

    def get_sender(self) -> str:
        raise NotImplementedError

    def get_source(self) -> str:
        raise NotImplementedError

    def get_now(self) -> int:
        raise NotImplementedError

    def get_level(self) -> int:
        raise NotImplementedError

    def get_balance(self) -> int:
        raise NotImplementedError

    def get_chain_id(self) -> str:
        raise NotImplementedError

    def get_dummy_address(self) -> str:
        raise NotImplementedError

    def get_dummy_txr_address(self) -> str:
        raise NotImplementedError

    def get_dummy_public_key(self) -> str:
        raise NotImplementedError

    def get_dummy_key_hash(self) -> str:
        raise NotImplementedError

    def get_dummy_signature(self) -> str:
        raise NotImplementedError

    def get_dummy_chain_id(self) -> str:
        raise NotImplementedError

    def get_dummy_lambda(self):
        raise NotImplementedError

    def set_total_voting_power(self, total_voting_power: int):
        raise NotImplementedError

    def set_voting_power(self, address: str, voting_power: int):
        raise NotImplementedError

    def get_operations_ttl(self) -> int:
        raise NotImplementedError

    def get_min_block_time(self) -> int:
        raise NotImplementedError



from typing import Callable
from typing import List
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from py_ecc import optimized_bls12_381 as bls12_381

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import dispatch_types
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BLS12_381_FrType
from pytezos.michelson.types import BLS12_381_G1Type
from pytezos.michelson.types import BLS12_381_G2Type
from pytezos.michelson.types import IntType
from pytezos.michelson.types import MutezType
from pytezos.michelson.types import NatType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import PairType
from pytezos.michelson.types import TimestampType
from pytezos.michelson.types.base import MichelsonType


class AbsInstruction(MichelsonInstruction, prim='ABS'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(IntType, stack.pop1())
        a.assert_type_equal(IntType)
        res = NatType.from_value(abs(int(a)))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)


class AddInstruction(MichelsonInstruction, prim='ADD'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(
            Tuple[
                Union[IntType, NatType, MutezType, TimestampType, BLS12_381_G1Type, BLS12_381_G2Type, BLS12_381_FrType],
                ...,
            ],
            stack.pop2(),
        )
        (res_type,) = dispatch_types(
            type(a),
            type(b),
            mapping={
                (NatType, NatType): (NatType,),
                (NatType, IntType): (IntType,),
                (IntType, NatType): (IntType,),
                (IntType, IntType): (IntType,),
                (TimestampType, IntType): (TimestampType,),
                (IntType, TimestampType): (TimestampType,),
                (MutezType, MutezType): (MutezType,),
                (BLS12_381_FrType, BLS12_381_FrType): (BLS12_381_FrType,),
                (BLS12_381_G1Type, BLS12_381_G1Type): (BLS12_381_G1Type,),
                (BLS12_381_G2Type, BLS12_381_G2Type): (BLS12_381_G2Type,),
            },
        )
        res_type = cast(
            Union[
                Type[IntType],
                Type[NatType],
                Type[TimestampType],
                Type[MutezType],
                Type[BLS12_381_G1Type],
                Type[BLS12_381_G2Type],
                Type[BLS12_381_FrType],
            ],
            res_type,
        )
        if issubclass(res_type, IntType):
            res = res_type.from_value(int(a) + int(b))  # type: ignore
        else:
            res = res_type.from_point(bls12_381.add(a.to_point(), b.to_point()))  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


class EdivInstruction(MichelsonInstruction, prim='EDIV'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(
            Tuple[Union[IntType, NatType, MutezType, TimestampType], Union[IntType, NatType, MutezType, TimestampType]],
            stack.pop2(),
        )
        q_type, r_type = dispatch_types(
            type(a),
            type(b),
            mapping={  # type: ignore
                (NatType, NatType): (NatType, NatType),
                (NatType, IntType): (IntType, NatType),
                (IntType, NatType): (IntType, NatType),
                (IntType, IntType): (IntType, NatType),
                (MutezType, NatType): (MutezType, MutezType),
                (MutezType, MutezType): (NatType, MutezType),
            },
        )  # type: Tuple[Union[Type[IntType], Type[NatType], Type[TimestampType], Type[MutezType]], Union[Type[IntType], Type[NatType], Type[TimestampType], Type[MutezType]]]
        if int(b) == 0:
            res = OptionType.none(PairType.create_type(args=[q_type, r_type]))
        else:
            q, r = divmod(int(a), int(b))
            if r < 0:
                r += abs(int(b))
                q += 1
            items: List[MichelsonType] = [q_type.from_value(q), r_type.from_value(r)]
            res = OptionType.from_some(PairType.from_comb(items))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


def execute_shift(prim: str, stack: MichelsonStack, stdout: List[str], shift: Callable[[Tuple[int, int]], int]):
    a, b = cast(Tuple[NatType, NatType], stack.pop2())
    a.assert_type_equal(NatType)
    b.assert_type_equal(NatType)
    assert int(b) < 257, f'shift overflow {int(b)}, should not exceed 256'
    c = shift((int(a), int(b)))
    res = NatType.from_value(c)
    stack.push(res)
    stdout.append(format_stdout(prim, [a, b], [res]))


class LslInstruction(MichelsonInstruction, prim='LSL'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_shift(cls.prim, stack, stdout, lambda x: x[0] << x[1])  # type: ignore
        return cls(stack_items_added=1)


class LsrInstruction(MichelsonInstruction, prim='LSR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_shift(cls.prim, stack, stdout, lambda x: x[0] >> x[1])  # type: ignore
        return cls(stack_items_added=1)


class MulInstruction(MichelsonInstruction, prim='MUL'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(
            Tuple[Union[IntType, NatType, MutezType, BLS12_381_FrType, BLS12_381_G1Type, BLS12_381_G2Type], ...],
            stack.pop2(),
        )
        (res_type,) = dispatch_types(
            type(a),
            type(b),
            mapping={
                (NatType, NatType): (NatType,),
                (NatType, IntType): (IntType,),
                (IntType, NatType): (IntType,),
                (IntType, IntType): (IntType,),
                (MutezType, NatType): (MutezType,),
                (NatType, MutezType): (MutezType,),
                (NatType, BLS12_381_FrType): (BLS12_381_FrType,),
                (IntType, BLS12_381_FrType): (BLS12_381_FrType,),
                (BLS12_381_FrType, NatType): (BLS12_381_FrType,),
                (BLS12_381_FrType, IntType): (BLS12_381_FrType,),
                (BLS12_381_FrType, BLS12_381_FrType): (BLS12_381_FrType,),
                (BLS12_381_G1Type, BLS12_381_FrType): (BLS12_381_G1Type,),
                (BLS12_381_G2Type, BLS12_381_FrType): (BLS12_381_G2Type,),
            },
        )
        res_type = cast(
            Union[
                Type[IntType],
                Type[NatType],
                Type[TimestampType],
                Type[MutezType],
                Type[BLS12_381_FrType],
                Type[BLS12_381_G1Type],
                Type[BLS12_381_G2Type],
            ],
            res_type,
        )
        if issubclass(res_type, IntType):
            res = res_type.from_value(int(a) * int(b))  # type: ignore
        else:
            res = res_type.from_point(bls12_381.multiply(a.to_point(), int(b)))  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NegInstruction(MichelsonInstruction, prim='NEG'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(Union[IntType, NatType, BLS12_381_FrType, BLS12_381_G1Type, BLS12_381_G2Type], stack.pop1())
        (res_type,) = dispatch_types(
            type(a),
            mapping={
                (IntType,): (IntType,),
                (NatType,): (IntType,),
                (BLS12_381_FrType,): (BLS12_381_FrType,),
                (BLS12_381_G1Type,): (BLS12_381_G1Type,),
                (BLS12_381_G2Type,): (BLS12_381_G2Type,),
            },
        )
        if issubclass(res_type, IntType):
            res = IntType.from_value(-int(a))  # type: ignore
        else:
            res = res_type.from_point(bls12_381.neg(a.to_point()))  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SubInstruction(MichelsonInstruction, prim='SUB'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(Tuple[Union[IntType, NatType, MutezType, TimestampType], ...], stack.pop2())
        (res_type,) = dispatch_types(
            type(a),
            type(b),
            mapping={  # type: ignore
                (NatType, NatType): (IntType,),
                (NatType, IntType): (IntType,),
                (IntType, NatType): (IntType,),
                (IntType, IntType): (IntType,),
                (TimestampType, IntType): (TimestampType,),
                (TimestampType, TimestampType): (IntType,),
                (MutezType, MutezType): (MutezType,),
            },
        )  # type: Tuple[Union[Type[IntType], Type[NatType], Type[TimestampType], Type[MutezType]]]
        res = res_type.from_value(int(a) - int(b))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SubMutezInstruction(MichelsonInstruction, prim='SUB_MUTEZ'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(Tuple[MutezType, MutezType], stack.pop2())
        a.assert_type_equal(MutezType)
        b.assert_type_equal(MutezType)
        try:
            res = OptionType.from_some(MutezType.from_value(int(a) - int(b)))
        except OverflowError:
            res = OptionType.none(MutezType)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


class IntInstruction(MichelsonInstruction, prim='INT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(Union[NatType, BLS12_381_FrType], stack.pop1())
        a.assert_type_in(NatType, BLS12_381_FrType)
        res = IntType.from_value(int(a))
        stack.push(res)
        stdout.append(f'{cls.prim} / {repr(a)} => {repr(res)}')
        return cls(stack_items_added=1)


class IsNatInstruction(MichelsonInstruction, prim='ISNAT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(IntType, stack.pop1())
        a.assert_type_equal(IntType)
        if int(a) >= 0:
            res = OptionType.from_some(NatType.from_value(int(a)))
        else:
            res = OptionType.none(NatType)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)



import json
from datetime import datetime

from pytezos.logging import logger

line_size = 100


def format_timestamp(timestamp: int) -> str:
    """Format unix timestamp.

    :param timestamp: Unix timestamp (seconds)
    """
    dt = datetime.utcfromtimestamp(timestamp)
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


class MichelsonFormatterError(ValueError):
    pass


def is_framed(node):
    if node['prim'] in {
        'Pair',
        'Left',
        'Right',
        'Some',
        'pair',
        'or',
        'option',
        'map',
        'big_map',
        'list',
        'set',
        'contract',
        'lambda',
        'ticket',
        'sapling_state',
        'sapling_transaction',
        'sapling_transaction_deprecated',
    }:
        return True
    elif node['prim'] in {
        'key',
        'unit',
        'signature',
        'operation',
        'int',
        'nat',
        'string',
        'bytes',
        'mutez',
        'bool',
        'key_hash',
        'timestamp',
        'address',
        'bls12_381_g1',
        'bls12_381_g2',
        'bls12_381_fr',
        'chain_id',
        'never',
    }:
        return 'annots' in node
    return False


def is_complex(node):
    return node['prim'] == 'LAMBDA' or node['prim'].startswith('IF')


def is_inline(node):
    return node['prim'] == 'PUSH'


def is_script(node):
    return all(
        map(
            lambda x: isinstance(x, dict)
            and x.get('prim')
            in [
                'parameter',
                'storage',
                'code',
            ],
            node,
        )
    )


def format_node(node, indent='', inline=False, is_root=False, wrapped=False):
    if isinstance(node, list):
        is_script_root = is_root and is_script(node)
        seq_indent = indent if is_script_root else indent + ' ' * 2
        items = list(map(lambda x: format_node(x, seq_indent, inline, wrapped=True), node))
        if items:
            length = len(indent) + sum(map(len, items)) + 4
            space = '' if is_script_root else ' '
            if inline or length < line_size:
                seq = f'{space}; '.join(items)
            else:
                seq = f'{space};\n{seq_indent}'.join(items)
            return seq if is_script_root else f'{{ {seq} }}'
        else:
            return '{}'

    elif isinstance(node, dict):
        if node.get('prim'):
            expr = ' '.join([node['prim']] + node.get('annots', []))
            args = node.get('args', [])

            if is_complex(node):
                arg_indent = indent + ' ' * 2
                items = list(map(lambda x: format_node(x, arg_indent, inline), args))
                length = len(indent) + len(expr) + sum(map(len, items)) + len(items) + 1
                if inline or length < line_size:
                    expr = f'{expr} {" ".join(items)}'
                else:
                    expr = f'\n{arg_indent}'.join([expr] + items)

            elif len(args) == 1:
                arg_indent = indent + ' ' * (len(expr) + 1)
                expr = f'{expr} {format_node(args[0], arg_indent, inline)}'

            elif len(args) > 1:
                arg_indent = indent + ' ' * 2
                alt_indent = indent + ' ' * (len(expr) + 2)
                for arg in args:
                    item = format_node(arg, arg_indent, inline)
                    length = len(indent) + len(expr) + len(item) + 1
                    if inline or is_inline(node) or length < line_size:
                        arg_indent = alt_indent
                        expr = f'{expr} {item}'
                    else:
                        expr = f'{expr}\n{arg_indent}{item}'

            if is_framed(node) and not is_root and not wrapped:
                return f'({expr})'
            else:
                return expr
        else:
            core_type, value = next((k, v) for k, v in node.items() if k[0] != '_' and k != 'annots')
            if core_type == 'int':
                return value
            elif core_type == 'bytes':
                return f'0x{value}'
            elif core_type == 'string':
                return json.dumps(value)
            else:
                raise AssertionError(f'unexpected core node {node}')
    else:
        raise AssertionError(f'unexpected node {node}')


def micheline_to_michelson(data, inline=False, wrap=False) -> str:
    """Converts micheline expression into formatted Michelson source.

    :param data: Micheline expression
    :param inline: produce single line, used for tezos-client arguments (False by default)
    :param wrap: ensure expression is wrapped in brackets
    """
    try:
        res = format_node(data, inline=inline, is_root=True)
        if wrap and any(map(res.startswith, ['Left', 'Right', 'Some', 'Pair'])):
            return f'({res})'
        else:
            return res
    except (KeyError, IndexError, TypeError) as e:
        logger.info(data)
        raise MichelsonFormatterError(e.args)



from traceback import format_exception
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import traitlets
from ipykernel.kernelbase import Kernel
from tabulate import tabulate

from michelson_kernel import __version__
from michelson_kernel.docs import docs
from pytezos import micheline_to_michelson
from pytezos.michelson.instructions import BigMapDiffInstruction
from pytezos.michelson.instructions import CommitInstruction
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.jupyter import RunInstruction
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.parse import MichelsonParserError
from pytezos.michelson.repl import Interpreter
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.tags import prim_tags
from pytezos.michelson.types import OperationType
from pytezos.michelson.types import PairType
from pytezos.michelson.types.domain import LambdaType
from pytezos.michelson.types.list import ListType
from pytezos.michelson.types.map import MapType
from pytezos.michelson.types.option import OptionType
from pytezos.michelson.types.set import SetType

static_macros = [
    'CMPEQ',
    'CMPNEQ',
    'CMPLT',
    'CMPGT',
    'CMPLE',
    'CMPGE',
    'IFEQ',
    'IFNEQ',
    'IFLT',
    'IFGT',
    'IFLE',
    'IFGE',
    'IFCMPEQ',
    'IFCMPNEQ',
    'IFCMPLT',
    'IFCMPGT',
    'IFCMPLE',
    'IFCMPGE',
    'FAIL',
    'ASSERT_EQ',
    'ASSERT_NEQ',
    'ASSERT_LT',
    'ASSERT_GT',
    'ASSERT_LE',
    'ASSERT_GE',
    'ASSERT_CMPEQ',
    'ASSERT_CMPNEQ',
    'ASSERT_CMPLT',
    'ASSERT_CMPGT',
    'ASSERT_CMPLE',
    'ASSERT_CMPGE',
    'ASSERT_NONE',
    'ASSERT_SOME',
    'ASSERT_LEFT',
    'ASSERT_RIGHT',
    'UNPAIR',
    'IF_SOME',
    'SET_CAR',
    'SET_CDR',
    'MAP_CAR',
    'MAP_CDR',
]


def parse_token(line: str, cursor_pos: int) -> Tuple[str, int, int]:

    begin_pos = next((i + 1 for i in range(cursor_pos - 1, 0, -1) if line[i] in ' ;({\n'), 0)
    end_pos = next((i for i in range(cursor_pos, len(line)) if line[i] in ' ;){\n'), len(line))
    return line[begin_pos:end_pos], begin_pos, end_pos


def format_prim(item: Union[MichelsonInstruction, Type[MichelsonInstruction]]) -> str:
    """Recursively get type hint for given type"""
    if isinstance(item, PairType) and item.items:
        return f'{item.prim} ({" ".join(format_prim(i) for i in item.args)})'
    if isinstance(item, OptionType) and item.item:
        return f'{item.prim} ({item.item.prim})'
    if isinstance(item, SetType) and item.items:
        return f'{item.prim} ({" ".join(format_prim(i) for i in item.args)})'
    if isinstance(item, ListType):
        return f'{item.prim} ({" ".join(format_prim(i) for i in item.args)})'
    if isinstance(item, MapType):
        return f'{item.prim} ({" ".join(format_prim(i) for i in item.args)})'
    if isinstance(item, type) and issubclass(item, LambdaType):
        return f'{item.prim} ({" ".join(format_prim(i) for i in item.args)})'
    return cast(str, item.prim)


def preformat_operations_table(items: List[OperationType]) -> List[Dict[str, Any]]:
    return [
        {
            'type': item.prim,
            **item.content,
        }
        for item in items
    ]


def preformat_storage_table(item) -> List[Dict[str, Any]]:
    prim = format_prim(item)
    value = item.to_python_object()
    if isinstance(value, bytes):
        value = value.hex()
    return [
        {
            'type': prim,
            'value': value,
        }
    ]


def preformat_stack_table(items: List[MichelsonInstruction]) -> List[Dict[str, Any]]:
    rows = []
    for i, item in enumerate(items):
        prim = format_prim(item)
        value = item.to_python_object()  # type: ignore
        if isinstance(value, bytes):
            value = value.hex()

        rows.append(
            {
                'index': i,
                'type': prim,
                'value': value,  # type: ignore
            }
        )
    return rows


def html_table(table: List[Dict[str, Any]], header: str) -> str:
    def pre(s):
        return f'<pre style="text-align: left;">{s}</pre>'

    def pre_dict(d):
        return {k: pre(v) for k, v in d.items()}

    result = f'<h4>{header}</h4>'
    result += tabulate(list(map(pre_dict, table)), tablefmt='html', headers="keys")
    # NOTE: Tabulate escapes our <pre> tags
    result = result.replace('&lt;', '<').replace('&gt;', '>')
    return result


def plain_table(table: List[Dict[str, Any]], header: str) -> str:
    result = f'{header}\n'
    return result + tabulate(table, tablefmt='simple', headers='keys')


def preformat_lazy_diff_table(lazy_diff: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    table = []
    for diff in lazy_diff:
        _id = diff['id']
        diff = diff['diff']
        if diff['action'] == 'alloc':
            table.append(
                {
                    'id': _id,
                    'action': 'alloc',
                    'key': '',
                    'value': '',
                }
            )
        for update in diff['updates']:
            table.append(
                {
                    'id': _id,
                    'action': 'update',
                    'key': micheline_to_michelson(update['key']),
                    'value': micheline_to_michelson(update['value']),
                }
            )
    return table


class MichelsonKernel(Kernel):
    implementation = 'IMichelson'
    implementation_version = __version__
    language_info = {
        'name': 'Michelson',
        'mimetype': 'text/x-michelson',
        'file_extension': '.tz',
        'codemirror_mode': 'michelson',
    }
    banner = 'Michelson (Tezos VM language)'
    help_links = traitlets.List(
        [
            'https://michelson.nomadic-labs.com/',
            'https://tezos.gitlab.io/whitedoc/michelson.html',
        ]
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.interpreter = Interpreter()

    def _stdout(self, text: str) -> None:
        self.send_response(
            self.iopub_socket,
            'stream',
            {
                'name': 'stdout',
                'text': text,
            },
        )

    def _find_stack_items(
        self,
        instructions: MichelineSequence,
        stack: MichelsonStack,
    ) -> Optional[List[MichelsonInstruction]]:
        for operation in instructions.items[::-1]:
            items = getattr(operation, 'items', None)
            if isinstance(items, list):
                stack_items = self._find_stack_items(MichelineSequence(items), stack)
                if stack_items:
                    return stack_items
            if not isinstance(operation, MichelsonInstruction):
                continue
            if operation.stack_items_added:
                return cast(List[MichelsonInstruction], stack.items[-operation.stack_items_added :])
        return None

    def _find_lazy_diff(self, instructions: MichelineSequence) -> Optional[List[Dict[str, str]]]:
        for instruction in instructions.items[::-1]:
            if isinstance(instruction, CommitInstruction) and instruction.lazy_diff:
                return instruction.lazy_diff
            if isinstance(instruction, BigMapDiffInstruction) and instruction.lazy_diff:
                return instruction.lazy_diff
        return None

    def _find_contract_result(self, instructions: MichelineSequence) -> Optional[PairType]:
        for instruction in instructions.items[::-1]:
            if isinstance(instruction, MichelineSequence):
                return self._find_contract_result(instruction)
            if isinstance(
                instruction,
                (
                    CommitInstruction,
                    RunInstruction,
                ),
            ):
                return instruction.result
        return None

    def _send_success_response(self, instructions: MichelineSequence, stack: MichelsonStack) -> Dict[str, Any]:
        plain, html = '', ''

        contract_result = self._find_contract_result(instructions)
        if contract_result:
            header = 'Operations'
            table = preformat_operations_table(cast(List[OperationType], contract_result.items[0]))
            plain += plain_table(table, header)
            html += html_table(table, header)

            header = 'Storage'
            table = preformat_storage_table(contract_result.items[1])
            plain += plain_table(table, header)
            html += html_table(table, header)

        modified_items = self._find_stack_items(instructions, stack)
        if modified_items is not None and not contract_result:
            header = 'Stack updates'
            table = preformat_stack_table(modified_items)
            plain += plain_table(table, header)
            html += html_table(table, header)

        lazy_diff = self._find_lazy_diff(instructions)
        if lazy_diff:
            header = 'BigMap diff'
            lazy_diff_table = preformat_lazy_diff_table(lazy_diff)
            plain += plain_table(lazy_diff_table, header)
            html += html_table(lazy_diff_table, header)

        result = {
            'data': {
                'text/plain': plain,
                'text/html': html,
            },
            'metadata': {},
            'execution_count': self.execution_count,
        }

        self.send_response(
            self.iopub_socket,
            'execute_result',
            result,
        )
        return result

    def _send_fail_response(self, error: Exception) -> Dict[str, Any]:
        if isinstance(error, (MichelsonParserError, MichelsonRuntimeError)):
            traceback = [error.format_stdout()]
        else:
            traceback = format_exception(error.__class__, error, None)
        result = {
            'status': 'error',
            'ename': error.__class__.__name__,
            'evalue': str(error),
            'traceback': traceback,
        }
        self.send_response(
            self.iopub_socket,
            'stream',
            {
                'name': 'stderr',
                'text': '\n'.join(traceback),
            },
        )
        return result

    def do_execute(
        self,
        code,
        silent,
        store_history=True,
        user_expressions=None,
        allow_stdin=False,
        *,
        cell_id=None,
    ):

        interpreter_result = self.interpreter.execute(code)

        if not silent and interpreter_result.stdout:
            self._stdout('\n'.join(interpreter_result.stdout))

        if not interpreter_result.error:
            if not silent:
                return self._send_success_response(interpreter_result.instructions, interpreter_result.stack)
        else:
            return self._send_fail_response(interpreter_result.error)
        return {}

    def do_complete(self, code, cursor_pos):
        token, begin_pos, end_pos = parse_token(code, cursor_pos)

        suggests = []
        for word_set in [prim_tags, static_macros]:
            for word in word_set:
                if word.startswith(token):
                    suggests.append(word)

        if suggests:
            res = {
                'matches': suggests,
                'cursor_start': begin_pos,
                'cursor_end': end_pos,
            }
        else:
            res = {
                'matches': [],
                'cursor_start': cursor_pos,
                'cursor_end': cursor_pos,
            }

        res['status'] = 'ok'
        return res

    def do_inspect(
        self,
        code,
        cursor_pos,
        detail_level=0,
        omit_sections=(),
    ):
        token, _, _ = parse_token(code, cursor_pos)
        docstring = docs.get(token)
        if docstring:
            res = {'found': True, 'data': {'text/plain': docstring}}
        else:
            res = {'found': False}

        res['status'] = 'ok'
        return res



from pprint import pformat
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from pytezos.context.mixin import ContextMixin
from pytezos.context.mixin import ExecutionContext
from pytezos.contract.call import ContractCall
from pytezos.jupyter import get_class_docstring
from pytezos.logging import logger
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.sections.parameter import ParameterSection


class ContractEntrypoint(ContextMixin):
    """Proxy class for spawning ContractCall instances."""

    def __init__(self, context: ExecutionContext, entrypoint: str) -> None:
        super().__init__(context=context)
        self.entrypoint = entrypoint

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            f'.entrypoint\t{self.entrypoint}',
            f'\nBuiltin\n(*args, **kwargs)\t# build transaction parameters (see typedef)',
            f'\nTypedef\n{self.__doc__}',
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def __call__(self, *args, **kwargs) -> 'ContractCall':
        """Spawn a contract call proxy initialized with the entrypoint name

        :param args: entrypoint args
        :param kwargs: entrypoint key-value args
        :rtype: ContractCall
        """
        if args:
            if len(args) == 1:
                py_obj = args[0]
            else:
                py_obj = args
        elif kwargs:
            py_obj = kwargs
        else:
            py_obj = None

        return ContractCall(
            context=self.context,
            parameters=self.encode(py_obj, self.context.mode),
        )

    def decode(self, value: Union[str, Dict[str, Any]], entrypoint: Optional[str] = None) -> Dict[str, Any]:
        """Convert from Michelson to Python type system

        :param value: Micheline JSON expression or Michelson value
        :param entrypoint: overwrite current entrypoint (in case you want to parse tx parameters)
        :return: Python object {entrypoint: value}
        """
        if isinstance(value, str):
            value = michelson_to_micheline(value)
        if entrypoint is None:
            entrypoint = self.entrypoint
        param_ty = ParameterSection.match(self.context.parameter_expr)
        parameters = {'entrypoint': entrypoint, 'value': value}
        py_obj = param_ty.from_parameters(parameters).to_python_object()
        return py_obj

    def encode(self, py_obj, mode: Optional[str] = None) -> dict:
        """Encode transaction parameters from the given Python object

        :param py_obj: Python object
        :param mode: whether to use `readable` or `optimized` (or `legacy_optimized`) encoding
        :return: {entrypoint, value}
        """
        try:
            param_ty = ParameterSection.match(self.context.parameter_expr)
            return param_ty.from_python_object({self.entrypoint: py_obj}).to_parameters(mode=mode or self.context.mode)
        except MichelsonRuntimeError as e:
            logger.info(self.__doc__)
            raise ValueError(f'Unexpected arguments: {pformat(py_obj)}', *e.args) from e



"""
Welcome to PyTezos!

To start playing with the Tezos blockchain you need to get a PyTezosClient instance.
Just type:

>>> from pytezos import pytezos
>>> pytezos

And follow the interactive documentation.
"""

from pytezos.client import PyTezosClient
from pytezos.contract.interface import Contract
from pytezos.contract.interface import ContractInterface
from pytezos.crypto.key import Key
from pytezos.logging import logger
from pytezos.michelson.forge import forge_micheline
from pytezos.michelson.forge import unforge_micheline
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import Undefined
from pytezos.michelson.types.core import Unit

__version__ = '3.7.1'

pytezos = PyTezosClient()



from typing import Generator
from typing import List
from typing import Tuple
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.types.base import MichelsonType


class ListType(MichelsonType, prim='list', args_len=1):
    def __init__(self, items: List[MichelsonType]):
        super(ListType, self).__init__()
        self.items = items

    def __repr__(self):
        return f'[{", ".join(map(repr, self.items))}]'

    def __len__(self):
        return len(self.items)

    def __iter__(self) -> Generator[MichelsonType, None, None]:
        yield from self.items

    def __eq__(self, other) -> bool:
        if not isinstance(other, ListType):
            return False
        return self.items == other.items

    @staticmethod
    def empty(item_type: Type[MichelsonType]):
        cls = ListType.create_type(args=[item_type])
        return cls(items=[])  # type: ignore

    @staticmethod
    def from_items(items: List[MichelsonType]):
        assert len(items) > 0, 'cannot instantiate from empty list'
        item_type = type(items[0])
        for item in items[1:]:
            item_type.assert_type_equal(item.get_anon_type())
        cls = ListType.create_type(args=[item_type])
        return cls(items)  # type: ignore

    @classmethod
    def generate_pydoc(cls, definitions: List[Tuple[str, str]], inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name
        arg_doc = cls.args[0].generate_pydoc(definitions, inferred_name=f'{name}_item' if name else None)
        return f'[ {arg_doc}, … ]'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'ListType':
        return cls([])

    @classmethod
    def from_micheline_value(cls, val_expr: list) -> 'ListType':
        assert isinstance(val_expr, list), f'expected list, got {type(val_expr).__name__}'
        items = list(map(cls.args[0].from_micheline_value, val_expr))
        return cls(items)

    @classmethod
    def from_python_object(cls, py_obj) -> 'ListType':
        assert isinstance(py_obj, list), f'expected list, got {type(py_obj).__name__}'
        items = list(map(cls.args[0].from_python_object, py_obj))
        return cls(items)

    def to_literal(self) -> Type[Micheline]:
        return MichelineSequence.create_type(args=[item.to_literal() for item in self.items])

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return list(map(lambda x: x.to_micheline_value(mode=mode, lazy_diff=lazy_diff), self))

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'list is not comparable'
        return list(
            map(
                lambda x: x.to_python_object(
                    try_unpack=try_unpack,
                    lazy_diff=lazy_diff,
                ),
                self,
            )
        )

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'MichelsonType':
        items = [item.merge_lazy_diff(lazy_diff) for item in self]
        return type(self)(items)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable'):
        items = [item.aggregate_lazy_diff(lazy_diff, mode=mode) for item in self]
        return type(self)(items)

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        for item in self:
            item.attach_context(context, big_map_copy=big_map_copy)

    def split_head(self) -> Tuple[MichelsonType, 'ListType']:
        assert len(self) > 0, f'cannot split empty list'
        head = self.items[0]
        tail = type(self)(self.items[1:])
        return head, tail

    def prepend(self, item: MichelsonType) -> 'ListType':
        self.args[0].assert_type_equal(type(item))
        return type(self)([item] + self.items)

    def __getitem__(self, idx: int) -> MichelsonType:
        assert isinstance(idx, int), f'expected int, got {type(idx).__name__}'
        assert idx < len(self), f'index out of bounds: {idx} >= {len(self)}'
        return self.items[idx]



from typing import Union

import base58


def tb(l):
    return b''.join(map(lambda x: x.to_bytes(1, 'big'), l))


base58_encodings = [
    #    Encoded   |               Decoded             |
    # prefix | len | prefix                      | len | Data type
    (b"B", 51, tb([1, 52]), 32, "block hash"),
    (b"o", 51, tb([5, 116]), 32, "operation hash"),
    (b"Lo", 52, tb([133, 233]), 32, "operation list hash"),
    (b"LLo", 53, tb([29, 159, 109]), 32, "operation list list hash"),
    (b"P", 51, tb([2, 170]), 32, "protocol hash"),
    (b"Co", 52, tb([79, 199]), 32, "context hash"),
    (b"tz1", 36, tb([6, 161, 159]), 20, "ed25519 public key hash"),
    (b"tz2", 36, tb([6, 161, 161]), 20, "secp256k1 public key hash"),
    (b"tz3", 36, tb([6, 161, 164]), 20, "p256 public key hash"),
    (b"KT1", 36, tb([2, 90, 121]), 20, "originated address"),
    # FIXME: replace with tb()
    (b"txr1", 37, b'\x01\x80x\x1f', 20, "tx_rollup_l2_address"),
    (b"id", 30, tb([153, 103]), 16, "cryptobox public key hash"),
    (b'expr', 54, tb([13, 44, 64, 27]), 32, u'script expression'),
    (b"edsk", 54, tb([13, 15, 58, 7]), 32, "ed25519 seed"),
    (b"edpk", 54, tb([13, 15, 37, 217]), 32, "ed25519 public key"),
    (b"spsk", 54, tb([17, 162, 224, 201]), 32, "secp256k1 secret key"),
    (b"p2sk", 54, tb([16, 81, 238, 189]), 32, "p256 secret key"),
    (b"edesk", 88, tb([7, 90, 60, 179, 41]), 56, "ed25519 encrypted seed"),
    (b"spesk", 88, tb([9, 237, 241, 174, 150]), 56, "secp256k1 encrypted secret key"),
    (b"p2esk", 88, tb([9, 48, 57, 115, 171]), 56, "p256_encrypted_secret_key"),
    (b"sppk", 55, tb([3, 254, 226, 86]), 33, "secp256k1 public key"),
    (b"p2pk", 55, tb([3, 178, 139, 127]), 33, "p256 public key"),
    (b"SSp", 53, tb([38, 248, 136]), 33, "secp256k1 scalar"),
    (b"GSp", 53, tb([5, 92, 0]), 33, "secp256k1 element"),
    (b"edsk", 98, tb([43, 246, 78, 7]), 64, "ed25519 secret key"),
    (b"edsig", 99, tb([9, 245, 205, 134, 18]), 64, "ed25519 signature"),
    (b"spsig", 99, tb([13, 115, 101, 19, 63]), 64, "secp256k1 signature"),
    (b"p2sig", 98, tb([54, 240, 44, 52]), 64, "p256 signature"),
    (b"sig", 96, tb([4, 130, 43]), 64, "generic signature"),
    (b'Net', 15, tb([87, 82, 0]), 4, "chain id"),
    (b'nce', 53, tb([69, 220, 169]), 32, 'seed nonce hash'),
    (b'btz1', 37, tb([1, 2, 49, 223]), 20, 'blinded public key hash'),
    (b'vh', 52, tb([1, 106, 242]), 32, 'block_payload_hash'),
]

operation_tags = {
    'endorsement': 0,
    'seed_nonce_revelation': 1,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 3,
    'account_activation': 4,
    'proposals': 5,
    'ballot': 6,
    'reveal': 7,
    'transaction': 8,
    'origination': 9,
    'delegation': 10,
}


def scrub_input(v: Union[str, bytes]) -> bytes:
    if isinstance(v, bytes):
        pass
    elif isinstance(v, str):
        try:
            _ = int(v, 16)
        except ValueError:
            v = v.encode('ascii')
        else:
            if v.startswith('0x'):
                v = v[2:]
            v = bytes.fromhex(v)
    else:
        raise TypeError("a bytes-like object is required (also str), not '%s'" % type(v).__name__)
    return v


def base58_decode(v: bytes) -> bytes:
    """Decode data using Base58 with checksum + validate binary prefix against known kinds and cut in the end.

    :param v: Array of bytes (use string.encode())
    :returns: bytes
    """
    try:
        prefix_len = next(
            len(encoding[2]) for encoding in base58_encodings if len(v) == encoding[1] and v.startswith(encoding[0])
        )
    except StopIteration as e:
        raise ValueError('Invalid encoding, prefix or length mismatch.') from e

    return base58.b58decode_check(v)[prefix_len:]


def base58_encode(v: bytes, prefix: bytes) -> bytes:
    """Encode data using Base58 with checksum and add an according binary prefix in the end.

    :param v: Array of bytes
    :param prefix: Human-readable prefix (use b'') e.g. b'tz', b'KT', etc
    :returns: bytes (use string.decode())
    """
    try:
        encoding = next(encoding for encoding in base58_encodings if len(v) == encoding[3] and prefix == encoding[0])
    except StopIteration as e:
        raise ValueError('Invalid encoding, prefix or length mismatch.') from e
    return base58.b58encode_check(encoding[2] + v)


def _validate(v: Union[str, bytes], prefixes: list):
    if isinstance(v, str):
        v = v.encode()
    v = scrub_input(v)
    if any(map(v.startswith, prefixes)):
        base58_decode(v)
    else:
        raise ValueError('Unknown prefix.')


def validate_pkh(v: Union[str, bytes]):
    """Ensure parameter is a public key hash (starts with b'tz1', b'tz2', b'tz3')

    :param v: string or bytes
    :raises ValueError: if parameter is not a public key hash
    """
    return _validate(v, prefixes=[b'tz1', b'tz2', b'tz3'])


def validate_l2_pkh(v: Union[str, bytes]):
    """Ensure parameter is a L2 public key hash (starts with b'txr1')

    :param v: string or bytes
    :raises ValueError: if parameter is not a public key hash
    """
    return _validate(v, prefixes=[b'txr1'])


def validate_sig(v: Union[str, bytes]):
    """Ensure parameter is a signature (starts with b'edsig', b'spsig', b'p2sig', b'sig')

    :param v: string or bytes
    :raises ValueError: if parameter is not a signature
    """
    return _validate(v, prefixes=[b'edsig', b'spsig', b'p2sig', b'sig'])


def is_pkh(v: Union[str, bytes]) -> bool:
    """Check if value is a public key hash."""
    try:
        validate_pkh(v)
    except (ValueError, TypeError):
        return False
    return True


def is_l2_pkh(v: Union[str, bytes]) -> bool:
    """Check if value is an L2 public key hash."""
    try:
        validate_l2_pkh(v)
    except (ValueError, TypeError):
        return False
    return True


def is_sig(v: Union[str, bytes]) -> bool:
    """Check if value is a signature."""
    try:
        validate_sig(v)
    except (ValueError, TypeError):
        return False
    return True


def is_bh(v: Union[str, bytes]) -> bool:
    """Check if value is a block hash."""
    try:
        _validate(v, prefixes=[b'B'])
    except (ValueError, TypeError):
        return False
    return True


def is_ogh(v) -> bool:
    """Check if value is an operation group hash."""
    try:
        _validate(v, prefixes=[b'o'])
    except (ValueError, TypeError):
        return False
    return True


def is_kt(v: Union[str, bytes]) -> bool:
    """Check if value is a KT address."""
    try:
        _validate(v, prefixes=[b'KT1'])
    except (ValueError, TypeError):
        return False
    return True


def is_public_key(v: Union[str, bytes]) -> bool:
    """Check if value is a public key."""
    try:
        _validate(v, prefixes=[b"edsk", b"edpk", b"spsk", b"p2sk", b"sppk", b"p2pk"])
    except (ValueError, TypeError):
        return False
    return True


def is_chain_id(v: Union[str, bytes]) -> bool:
    """Check if value is a chain id."""
    try:
        _validate(v, prefixes=[b'Net'])
    except (ValueError, TypeError):
        return False
    return True


def is_address(v: Union[str, bytes]) -> bool:
    """Check if value is a tz/KT address"""
    if isinstance(v, bytes):
        v = v.decode()
    address = v.split('%')[0]
    return is_kt(address) or is_pkh(address)


def is_txr_address(v: Union[str, bytes]) -> bool:
    """Check if value is a txr1 address"""
    if isinstance(v, bytes):
        v = v.decode()
    address = v.split('%')[0]
    return is_l2_pkh(address)
